import csv

from django.core.management.base import BaseCommand, CommandError

from kmuhelper.modules.main.models import Customer
from kmuhelper.utils import getfirstindex


class Command(BaseCommand):
    help = "Imports customers from a .csv file."

    def add_arguments(self, parser):
        parser.add_argument(
            "filepath", help="Absolute filepath to the .csv file", type=str
        )

    def handle(self, *args, **options):
        self.stdout.write(self.style.SUCCESS("Import gestartet!"))
        try:
            with open(options["filepath"], newline="", encoding="utf-8-sig") as f:
                reader = csv.reader(f)
                rows = list(reader)

            indexrow = rows.pop(0)

            ###
            index_id = getfirstindex(indexrow, ["Nummer", "ID"])

            index_email = getfirstindex(indexrow, ["Email", "E-Mail", "EMail", "EMAIL"])
            index_first_name = getfirstindex(indexrow, ["Vorname"])
            index_last_name = getfirstindex(indexrow, ["Name", "Nachname"])
            index_language = getfirstindex(indexrow, ["Sprache", "SPRACHE"])

            index_company = getfirstindex(indexrow, ["Firma"])
            index_strasse = getfirstindex(indexrow, ["Strasse"])
            index_city = getfirstindex(indexrow, ["Ort"])
            index_postcode = getfirstindex(indexrow, ["PLZ", "Plz", "Postleitzahl"])
            index_country = getfirstindex(indexrow, ["Land"])
            index_phone = getfirstindex(indexrow, ["Telefon", "Telefon1"])

            index_address_2 = getfirstindex(indexrow, ["Adresszeile2", "Zusatz"])

            index_note = getfirstindex(indexrow, ["Notiz"])
            index_web = getfirstindex(indexrow, ["Webseite", "Homepage"])

            ###

            for row in rows:
                objid = row[index_id] if index_id is not None else None

                email = row[index_email] if index_email is not None else ""
                first_name = (
                    row[index_first_name] if index_first_name is not None else ""
                )
                last_name = row[index_last_name] if index_last_name is not None else ""
                language = (
                    (
                        "de"
                        if row[index_language].upper() in ["D", "DE", "DEUTSCH"]
                        else (
                            "fr"
                            if row[index_language].upper() in ["F", "FR", "FRANZÖSISCH"]
                            else (
                                "it"
                                if row[index_language].upper()
                                in ["I", "IT", "ITALIENISCH"]
                                else (
                                    "en"
                                    if row[index_language].upper()
                                    in ["E", "EN", "ENGLISCH"]
                                    else "de"
                                )
                            )
                        )
                    )
                    if index_language is not None
                    else "de"
                )

                company = row[index_company] if index_company is not None else ""
                address_1 = row[index_strasse] if index_strasse is not None else ""
                address_2 = row[index_address_2] if index_address_2 is not None else ""
                city = row[index_city] if index_city is not None else ""
                postcode = row[index_postcode] if index_postcode is not None else ""
                country = row[index_country] if index_country is not None else ""
                phone = row[index_phone] if index_phone is not None else ""

                note = row[index_note] if index_note is not None else ""
                website = row[index_web] if index_web is not None else ""

                oldwithid = Customer.objects.filter(pk=objid).exists()
                oldwithemail = (
                    Customer.objects.filter(email=email).exists() if email else False
                )
                oldwithidandemail = Customer.objects.filter(
                    pk=objid, email=email
                ).exists()

                self.stdout.write(self.style.SUCCESS(""))

                if not oldwithid and not oldwithemail:
                    self.stdout.write(
                        self.style.SUCCESS(
                            "Kunden erstellen:"
                            + "\n  ID:         "
                            + str(objid)
                            + "\n  E-Mail:     "
                            + email
                            + "\n  Vorname:    "
                            + first_name
                            + "\n  Nachname:   "
                            + last_name
                            + "\n  Firma:      "
                            + company
                        )
                    )
                    Customer.objects.create(
                        pk=objid,
                        email=email,
                        first_name=first_name,
                        last_name=last_name,
                        company=company,
                        language=language,
                        addr_billing_first_name=first_name,
                        addr_billing_last_name=last_name,
                        addr_billing_company=company,
                        addr_billing_address_1=address_1,
                        addr_billing_address_2=address_2,
                        addr_billing_city=city,
                        # addr_billing_state = "",
                        addr_billing_postcode=postcode,
                        addr_billing_country=country,
                        addr_billing_email=email,
                        addr_billing_phone=phone,
                        addr_shipping_first_name=first_name,
                        addr_shipping_last_name=last_name,
                        addr_shipping_company=company,
                        addr_shipping_address_1=address_1,
                        addr_shipping_address_2=address_2,
                        addr_shipping_city=city,
                        # addr_shipping_state = "",
                        addr_shipping_postcode=postcode,
                        addr_shipping_country=country,
                        website=website,
                        note=note,
                    )
                elif oldwithidandemail:
                    self.stdout.write(
                        self.style.SUCCESS(
                            "Kunde existiert bereits!"
                            + "\n  ID:         "
                            + str(objid)
                            + "\n  E-Mail:     "
                            + email
                            + "\n  Vorname:    "
                            + first_name
                            + "\n  Nachname:   "
                            + last_name
                            + "\n  Firma:      "
                            + company
                        )
                    )
                elif oldwithemail and not oldwithid:
                    old = Customer.objects.get(email=email)
                    self.stdout.write(
                        self.style.SUCCESS(
                            "Kunde mit der E-Mail '"
                            + email
                            + "' existiert bereits, jedoch ist die ID unterschiedlich!"
                            + "\n  KMUHelper-Daten:"
                            + "\n    ID:       "
                            + str(old.pk)
                            + "\n    Vorname:  "
                            + old.first_name
                            + "\n    Nachname: "
                            + old.last_name
                            + "\n    Firma:    "
                            + old.addr_billing_company
                            + "\n  CSV Daten:"
                            + "\n    ID:       "
                            + str(objid)
                            + "\n    Vorname:  "
                            + first_name
                            + "\n    Nachname: "
                            + last_name
                            + "\n    Firma:    "
                            + company
                        )
                    )
                    if input("ID aus der CSV-Datei verwenden? [y/n] ").lower() in [
                        "y",
                        "yes",
                        "j",
                        "ja",
                        "aktualisieren",
                    ]:
                        oldid = old.pk
                        old.pk = int(objid)
                        old.save()
                        new = old
                        old = Customer.objects.get(objid=oldid)
                        new.combine_with = old
                        new.save()
                        self.stdout.write(self.style.SUCCESS("ID aktualisiert!"))
                elif not oldwithemail and oldwithid:
                    old = Customer.objects.get(pk=objid)
                    self.stdout.write(
                        self.style.SUCCESS(
                            "Kunde mit der ID '"
                            + str(objid)
                            + "' existiert bereits, jedoch ist die E-Mail unterschiedlich!"
                            + "\n  KMUHelper-Daten:"
                            + "\n    E-Mail:   "
                            + old.email
                            + "\n    Vorname:  "
                            + old.first_name
                            + "\n    Nachname: "
                            + old.last_name
                            + "\n    Firma:    "
                            + old.addr_billing_company
                            + "\n  CSV Daten:"
                            + "\n    E-Mail:   "
                            + email
                            + "\n    Vorname:  "
                            + first_name
                            + "\n    Nachname: "
                            + last_name
                            + "\n    Firma:    "
                            + company
                        )
                    )
                    if input("E-Mail aus der CSV-Datei verwenden? [y/n] ").lower() in [
                        "y",
                        "yes",
                        "j",
                        "ja",
                        "aktualisieren",
                    ]:
                        old.email = email
                        old.save()
                        self.stdout.write(self.style.SUCCESS("E-Mail aktualisiert!"))
                elif oldwithemail and oldwithid:
                    self.stdout.write(
                        self.style.SUCCESS(
                            "Irgendetwas stimmt nicht! Es besteht ein Kunde mit der E-Mail '"
                            + email
                            + "' aber ein Anderer mit der ID '"
                            + str(objid)
                            + "'."
                        )
                    )

        except FileNotFoundError:
            raise CommandError("Datei nicht gefunden!")



from django.utils.functional import lazy
from django.utils.html import format_html
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy

from kmuhelper.utils import faq

_ = gettext_lazy
_t = format_lazy
_h = lazy(format_html, str)

SETTINGS = {
    "wc-url": {
        "typ": "url",
        "name": _("Shop-URL"),
        "description": _(
            "URL des WooCommerce-/Wordpress-Shops, z.B. https://shop.example.com"
        ),
    },
    "wc-webhook-secret": {
        "typ": "char",
        "name": _("Webhook-Secret"),
        "description": _("Secret-Key für WooCommerce-Webhooks"),
    },
    "email-stock-warning-receiver": {
        "typ": "email",
        "name": _("E-Mail-Adresse für Warnungen zum Lagerbestand"),
        "description": _t(
            "{}\n\n{}\n\n{}",
            _(
                "Wenn durch den Import von Bestellungen der Lagerbestand-Soll-Wert unterschritten wird, wird eine "
                "Warnung an diese E-Mail-Adresse gesendet."
            ),
            _("Leer lassen, um diese E-Mails zu deaktivieren."),
            _(
                "Bemerkung: Entsteht die Unterschreitung durch eine manuelle Aktion, wird eine Warnung direkt "
                "angezeigt."
            ),
        ),
    },
    "email-show-buttons": {
        "typ": "bool",
        "name": _("E-Mail-Knöpfe anzeigen"),
        "description": _(
            "Aktivieren oder Deaktivieren der E-Mail-Knöpfe bei Bestellungen und Kunden."
        ),
    },
    "email-signature": {
        "typ": "text",
        "name": _("E-Mail-Signatur"),
        "description": _("Standardsignatur für E-Mails"),
    },
    "default-payment-conditions": {
        "typ": "char",
        "name": _("Standardzahlungskonditionen"),
        "description": _h(
            "{} {}\n\n{} -> {}",
            _("Standardwert der Zahlungskonditionen für neue Bestellungen."),
            _("Beispiele: 0:30 oder 2:10;0:30"),
            _("Mehr über Zahlungskonditionen erfahren"),
            faq("wie-funktionieren-zahlungskonditionen"),
        ),
    },
    "print-payment-conditions": {
        "typ": "bool",
        "name": _("Zahlungskonditionen drucken"),
        "description": _h(
            "{}\n\n{}\n\n{} -> {}",
            _(
                "Wenn aktiviert, werden die Zahlungskonditionen bei Rechnungen in menschenlesbarer Form unter dem "
                "Rechnungstotal gedruckt."
            ),
            _(
                "Die Zahlungskonditionen werden unabhängig von dieser Einstellung immer auch in maschinenlesbarer "
                "Form im QR-Code integriert."
            ),
            _("Mehr über Zahlungskonditionen erfahren"),
            faq("wie-funktionieren-zahlungskonditionen"),
        ),
    },
}

SETTINGS_FIELDSETS = [
    {
        "name": _("Zahlungskonditionen"),
        "fields": ["default-payment-conditions", "print-payment-conditions"],
        "classes": "wide",
    },
    {
        "name": _("E-Mails"),
        "fields": [
            "email-show-buttons",
            "email-signature",
            "email-stock-warning-receiver",
        ],
        "classes": "wide",
    },
]

SECRET_SETTINGS = {
    "wc-consumer_key": {"typ": "char"},
    "wc-consumer_secret": {"typ": "char"},
    "wc-url": {"typ": "url"},
}



# Generated by Django 5.0.4 on 2024-04-17 20:28

import django.core.validators
import kmuhelper.modules.main.models
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("admin", "0003_logentry_add_action_flag_choices"),
        ("kmuhelper", "0113_customer_woocommerce_deleted_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="AdminLogEntry",
            fields=[],
            options={
                "verbose_name": "Admin-Logeintrag",
                "verbose_name_plural": "Admin-Logeinträge",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("admin.logentry",),
        ),
        migrations.AlterField(
            model_name="order",
            name="payment_conditions",
            field=models.CharField(
                blank=True,
                default=kmuhelper.modules.main.models.default_payment_conditions,
                help_text='Skonto und Zahlungsfrist -> <a target="_blank" href="https://rafaelurben.github.io/django-kmuhelper/manual/faq#wie-funktionieren-zahlungskonditionen">FAQ</a>',
                max_length=25,
                null=True,
                validators=[
                    django.core.validators.RegexValidator(
                        "^([0-9]+(\\.[0-9]+)?:[0-9]+;)*0:[0-9]+$",
                        "Bitte benutze folgendes Format: 'p:d;p:d' - p = Skonto in %; d = Tage",
                    )
                ],
                verbose_name="Zahlungskonditionen",
            ),
        ),
    ]



from django.db import models
from django.http import JsonResponse
from django.views.decorators.csrf import csrf_exempt

from kmuhelper.utils import package_version, python_version, formatprice
from kmuhelper.modules.main.models import Order
from kmuhelper.modules.api.constants import ENDPOINT_NOT_FOUND, SUCCESSFULLY_CHANGED
from kmuhelper.modules.api.decorators import (
    require_object,
    api_read,
    api_write,
    api_readwrite,
)

#####


def not_found(request):
    """Return a json error message"""

    return ENDPOINT_NOT_FOUND


#####


@api_read()
def versions(request):
    """Get the installed versions of some dependencies, their newest versions available and the current python version."""

    context = {
        "versions": {
            "Python": {"current": python_version(), "latest": "-", "uptodate": None},
        }
    }

    packages = [
        "Django",
        "django-kmuhelper",
        "requests",
        "WooCommerce",
        "reportlab",
        "rich",
        "defusedxml",
    ]

    for package in packages:
        context["versions"][package] = package_version(package)
    return JsonResponse(context)


@api_read(["view_order"])
def orders_unpaid(request):
    """Get sums of currently unpaid orders."""

    sum_unsent = (
        Order.objects.filter(is_paid=False, is_shipped=False).aggregate(
            models.Sum("cached_sum")
        )["cached_sum__sum"]
        or 0
    )
    sum_sent = (
        Order.objects.filter(is_paid=False, is_shipped=True).aggregate(
            models.Sum("cached_sum")
        )["cached_sum__sum"]
        or 0
    )

    context = {
        "orders_unpaid_sum": {
            "unsent": formatprice(round(sum_unsent, 2)),
            "sent": formatprice(round(sum_sent, 2)),
            "all": formatprice(round(sum_sent + sum_unsent, 2)),
        }
    }
    return JsonResponse(context)


@csrf_exempt
@api_write(["change_order"])
@require_object(Order)
def orders_set_paid(request, obj):
    """Set an order as paid"""

    if not obj.is_paid:
        obj.is_paid = True
        if obj.is_shipped and obj.status == "pending":
            obj.status = "completed"
        obj.save()
    return SUCCESSFULLY_CHANGED



from requests import HTTPError, JSONDecodeError

from kmuhelper.modules.integrations.woocommerce.api._base import WC_BaseAPI


class WCGeneralAPI(WC_BaseAPI):
    """Manage connection with WooCommerce"""

    LOG_PREFIX = "[deep_pink4][KMUHelper WooCommerce][/] -"

    # System status

    def get_system_status(self):
        """Get system status from WooCommerce"""
        try:
            response = self.wcapi.get("system_status")
            response.raise_for_status()
            return response.json()
        except (HTTPError, JSONDecodeError) as e:
            self.log("Error while getting system status: ", e)
            return False

    def get_enabled_webhooks_topics(self, delivery_url) -> tuple[bool, list | None]:
        """Check if webhooks have been setup correctly

        Returns: (success, topics)"""
        try:
            response = self.wcapi.get("webhooks?status=active&per_page=100")
            response.raise_for_status()

            webhooks = response.json()

            topics = []
            for webhook in webhooks:
                if webhook["delivery_url"] == delivery_url:
                    topics.append(webhook["topic"])
            return True, topics
        except (HTTPError, JSONDecodeError) as e:
            self.log("Error while getting webhooks: ", e)
            return False, None



from rich import print

from kmuhelper import settings


def log(string, *args):
    print("[deep_pink4][KMUHelper startup.py][/] -", string, *args)


# Set up database settings

log("Started... DEBUG is", settings.DEBUG)

try:
    log("Setting up settings...")
    settings.setup_settings()
    log("Set up settings!")
except Exception as error:
    log("Error while setting up settings:", type(error), error)

log("Ended.")



from django.contrib import admin

from kmuhelper.modules.settings import models
from kmuhelper.overrides import CustomModelAdmin


@admin.register(models.Setting)
class SettingAdmin(CustomModelAdmin):
    list_display = ("name", "content_display")
    ordering = ("id",)

    search_fields = []

    readonly_fields = ["id", "name", "description"]

    HIDDEN = True

    def get_fields(self, request, obj=None):
        return ["description", f"content_{obj.typ}"]

    # Permissions

    NO_ADD = True
    NO_DELETE = True


#

modeladmins = [
    (models.Setting, SettingAdmin),
]



import kmuhelper.modules.main.models as models
from kmuhelper.modules.integrations.woocommerce.api._base import WC_BaseObjectAPI
from kmuhelper.modules.integrations.woocommerce.api._utils import preparestring


class WCProductCategoriesAPI(WC_BaseObjectAPI):
    LOG_PREFIX = "[deep_pink4][KMUHelper WooCommerce Product categories][/] -"

    MODEL = models.ProductCategory
    WC_OBJ_DOES_NOT_EXIST_CODE = "woocommerce_rest_term_invalid"
    WC_API_BASEURL = "products/categories"

    def update_object_from_data(self, db_obj, wc_obj: dict):
        db_obj.woocommerce_deleted = False

        db_obj.name = preparestring(wc_obj["name"])
        db_obj.description = preparestring(wc_obj["description"])
        db_obj.image_url = (wc_obj["image"]["src"]) if wc_obj["image"] else ""
        if wc_obj["parent"]:
            obj, created = models.ProductCategory.objects.get_or_create(
                woocommerceid=wc_obj["parent"]
            )
            if created:
                self.update_object_from_api(obj)
            db_obj.parent_category = obj
        db_obj.save()

        self.log("Category updated:", str(db_obj))

    def create_object_from_data(self, wc_obj: dict):
        db_obj = models.ProductCategory.objects.create(
            name=preparestring(wc_obj["name"]),
            description=preparestring(wc_obj["description"]),
            image_url=(wc_obj["image"]["src"]) if wc_obj["image"] else "",
            woocommerceid=wc_obj["id"],
        )

        self.log("Category created:", wc_obj["name"])
        return db_obj

    def _post_process_imported_objects(
        self, db_obj__wc_obj_list: list[tuple[object, dict]]
    ):
        for db_obj, wc_obj in db_obj__wc_obj_list:
            if wc_obj["parent"]:
                db_obj.parent_category = models.ProductCategory.objects.get(
                    woocommerceid=wc_obj["parent"]
                )
                db_obj.save()



from django import forms

from kmuhelper import constants


class PDFOrderForm(forms.Form):
    # Fieldsets

    fieldsets = [
        {"fields": ["preset"]},
        {"fields": ["title", "text", "language"], "name": "Text & Sprache"},
        {"fields": ["do_print"], "name": "Optionen"},
        {"fields": ["do_download"], "name": "Ausgabe"},
    ]

    # Fields

    preset = forms.ChoiceField(
        label="Vorlage",
        choices=(
            ("invoice", "Rechnung"),
            ("delivery-note", "Lieferschein"),
            ("payment-reminder", "Zahlungserinnerung"),
        ),
        required=True,
        help_text="Die Vorlage definiert das Layout und die Inhalte der PDF-Datei",
    )
    title = forms.CharField(
        label="Titel",
        required=False,
        max_length=32,
        help_text="Z. B. 'Rechnung' oder 'Lieferschein' - Leer lassen für Standardwert der Vorlage",
    )
    text = forms.CharField(
        label="Text",
        required=False,
        widget=forms.Textarea,
        help_text="Dieser Text wird unterhalb des Titels angezeigt - Leer lassen für Standardwert der Vorlage",
    )
    language = forms.ChoiceField(
        label="Sprache",
        choices=constants.LANGUAGES,
        required=True,
        help_text="Die Sprache, in der die PDF-Datei generiert werden soll",
    )

    do_print = forms.BooleanField(
        label="Druckversion?",
        required=False,
        help_text="Wenn aktiviert, wird die PDF-Datei ohne Schnittmarker generiert",
    )
    do_download = forms.BooleanField(
        label="Herunterladen?",
        required=False,
        help_text="Datei automatisch herunterladen (dieses Verhalten kann je nach Browser variieren)",
    )

    def get_url_params(self):
        result = f"?custom&preset={self.cleaned_data['preset']}&language={self.cleaned_data['language']}"
        if self.cleaned_data["do_print"]:
            result += "&print"
        if self.cleaned_data["do_download"]:
            result += "&download"
        return result

    def update_order_settings(self, order):
        order.pdf_title = self.cleaned_data["title"]
        order.pdf_text = self.cleaned_data["text"]
        order.save()



from django.contrib import admin

from kmuhelper.modules.app.models import (
    App_ToDo,
    App_Shipping,
    App_IncomingPayments,
    App_Stock,
    App_Arrival,
)
from kmuhelper.modules.main.admin import (
    NoteAdmin,
    OrderAdmin,
    SupplyAdmin,
    ProductAdmin,
)
from kmuhelper.overrides import CustomModelAdmin


#######


class App_AdminBase(CustomModelAdmin):
    HIDDEN = True


@admin.register(App_ToDo)
class App_ToDoAdmin(App_AdminBase, NoteAdmin):
    list_display = ["pkfill", "name", "description", "priority", "done", "created_at"]
    list_editable = ["priority", "done"]
    list_filter = ["priority"]

    ordering = ["-priority", "created_at"]


@admin.register(App_Shipping)
class App_ShippingAdmin(App_AdminBase, OrderAdmin):
    list_display = [
        "pkfill",
        "info",
        "status",
        "shipped_on",
        "is_shipped",
        "tracking_number",
        "linked_note_html",
    ]
    list_editable = ["shipped_on", "is_shipped", "status", "tracking_number"]
    list_filter = ["status", "is_paid"]

    ordering = ["is_paid", "-date"]

    actions = []


@admin.register(App_IncomingPayments)
class App_IncomingPaymentsAdmin(App_AdminBase, OrderAdmin):
    list_display = [
        "pkfill",
        "info",
        "status",
        "paid_on",
        "is_paid",
        "display_cached_sum",
        "display_payment_conditions",
        "linked_note_html",
    ]
    list_editable = ["paid_on", "is_paid", "status"]
    list_filter = ["status", "is_shipped", "payment_method"]

    ordering = [
        "status",
        "invoice_date",
    ]

    actions = []


@admin.register(App_Stock)
class App_StockAdmin(App_AdminBase, ProductAdmin):
    list_display = [
        "pkfill",
        "display_article_number",
        "clean_name",
        "stock_current",
        "stock_target",
        "display_current_price",
        "note",
        "linked_note_html",
    ]
    list_editable = ["stock_current", "stock_target"]

    actions = ["reset_stock"]


@admin.register(App_Arrival)
class App_ArrivalAdmin(App_AdminBase, SupplyAdmin):
    list_display = ["pkfill", "name", "date", "total_quantity", "linked_note_html"]
    list_filter = []


#


modeladmins = [
    (App_ToDo, App_ToDoAdmin),
    (App_Shipping, App_ShippingAdmin),
    (App_IncomingPayments, App_IncomingPaymentsAdmin),
    (App_Stock, App_StockAdmin),
    (App_Arrival, App_ArrivalAdmin),
]



from django import template
from django.apps import apps
from django.urls import reverse
from django.utils.html import format_html

from kmuhelper import settings, constants
from kmuhelper.modules import config
from kmuhelper.modules.integrations.woocommerce.utils import (
    is_connected as is_woocommerce_connected,
)

register = template.Library()

##########


@register.simple_tag(name="kmuhelper_woocommerce_connected")
def kmuhelper_woocommerce_connected():
    return is_woocommerce_connected()


@register.simple_tag(name="kmuhelper_has_module_permission", takes_context=True)
def kmuhelper_has_module_permission(context, module_name):
    return config.user_has_module_permission(context.get("user"), module_name)


@register.simple_tag(name="kmuhelper_email_show_buttons", takes_context=True)
def kmuhelper_email_show_buttons(context):
    show = bool(settings.get_db_setting("email-show-buttons"))
    return show and context.get("user").has_perm("kmuhelper.view_email")


@register.simple_tag(name="get_admin_model", takes_context=True)
def get_admin_model(context, model_name=None):
    model = config.get_model_from_context(context, model_name)
    return model


@register.simple_tag(name="get_admin_module", takes_context=True)
def get_admin_module(context, module_name=None):
    model = config.get_module_from_context(context, module_name)
    return model


@register.simple_tag(name="kmuhelper_branding", takes_context=True)
def kmuhelper_branding(context, module_name=None):
    module = config.get_module_from_context(context, module_name)

    url_home = reverse("kmuhelper:home")
    url_module = reverse(module.get("viewname"))

    return format_html(
        '<h1 id="site-name"><a target="_top" href="{}">{}</a> | <a id="app-link" target="_top" href="{}">{}</a></h1>',
        url_home,
        "KMUHelper",
        url_module,
        module.get("title"),
    )


@register.inclusion_tag("kmuhelper/_includes/pagechooser.html", takes_context=False)
def kmuhelper_pagechooser_from_model_list(model_list):
    griddata = []
    for model in model_list:
        try:
            dbmodel = apps.get_model("kmuhelper", model["object_name"])
        except LookupError:
            dbmodel = None
        griddata.append(
            {
                "title": getattr(dbmodel, "ADMIN_TITLE", "") or model["name"],
                "url": model["admin_url"],
                "subtitle": getattr(dbmodel, "ADMIN_DESCRIPTION", ""),
                "icon": getattr(
                    dbmodel, "ADMIN_ICON", "fa-solid fa-circle-exclamation"
                ),
            }
        )
    return {"griddata": [griddata], "pagechooserclass": "smallboxes"}


@register.inclusion_tag("kmuhelper/_includes/pagechooser.html", takes_context=False)
def kmuhelper_pagechooser(griddata, pagechooserclass=""):
    return {"griddata": griddata, "pagechooserclass": pagechooserclass}


@register.simple_tag(name="docs_url", takes_context=False)
def docs_url(path=""):
    return constants.URL_MANUAL + path



"""Methods to get settings form different places"""

from django.conf import settings as djangoconfig
from django.core.exceptions import ObjectDoesNotExist

from kmuhelper.modules.settings.models import Setting, SettingHidden
from kmuhelper.modules.settings.constants import SETTINGS, SECRET_SETTINGS

DEBUG = djangoconfig.DEBUG
AUTH_USER_MODEL = djangoconfig.AUTH_USER_MODEL


def setup_settings():
    """Setup the database settings"""

    for key, value in SETTINGS.items():
        Setting.objects.update_or_create(id=key, defaults={"typ": value["typ"]})

    for key, value in SECRET_SETTINGS.items():
        SettingHidden.objects.update_or_create(id=key, defaults={"typ": value["typ"]})


# File settings


def get_file_setting(name, default=None):
    """Get a setting from the settings.py file"""

    return getattr(djangoconfig, name, default)


def has_file_setting(name):
    """Check if a setting in the settings.py file has been set"""

    return hasattr(djangoconfig, name)


# Get db


def get_db_setting(settingid, default=None):
    """Get a setting from the 'Setting' model"""

    try:
        setting = Setting.objects.get(id=settingid)
        if setting.typ in ["char", "text"] and setting.content == "":
            return default
        return setting.content
    except ObjectDoesNotExist:
        return default


def get_secret_db_setting(settingid, default=None):
    """Get a setting from the 'SettingHidden' model"""

    try:
        return SettingHidden.objects.get(id=settingid).content
    except ObjectDoesNotExist:
        return default


# Set db


def set_db_setting(settingid, content):
    """Update a database setting"""

    try:
        obj = Setting.objects.get(id=settingid)
        obj.content = content
        obj.save()
        return True
    except ObjectDoesNotExist:
        return False


def set_secret_db_setting(settingid, content):
    """Update a secret database setting"""

    try:
        obj = SettingHidden.objects.get(id=settingid)
        obj.content = content
        obj.save()
        return True
    except ObjectDoesNotExist:
        return False



from functools import wraps

from django.contrib import messages
from django.contrib.auth.decorators import user_passes_test, PermissionDenied
from django.http import Http404
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.translation import gettext

import kmuhelper.modules.config as config
from kmuhelper.utils import render_error

_ = gettext


def confirm_action(action_message):
    """Decorator to show a confirm page where the user has to
    confirm an action before executing it."""

    def decorator(function):
        @wraps(function)
        def wrap(request, *args, **kwargs):
            if request.method == "POST":
                return function(request, *args, **kwargs)
            return render(
                request, "admin/kmuhelper/_confirm.html", {"action": action_message}
            )

        return wrap

    return decorator


def require_object(
    model,
    redirect_url=None,
    raise_404=False,
    show_errorpage=False,
    custom_response=None,
):
    """Decorator to only call the view if an object with the given id exists
    and automatically pass it instead of the id."""

    def decorator(function):
        @wraps(function)
        def wrap(request, object_id, *args, **kwargs):
            if model.objects.filter(pk=int(object_id)).exists():
                obj = model.objects.get(pk=int(object_id))
                return function(request, obj, *args, **kwargs)

            if custom_response:
                return custom_response

            messages.warning(
                request,
                _("%(name)s with ID %(id)s was not found!")
                % {"name": model._meta.verbose_name, "id": object_id},
            )

            if raise_404:
                raise Http404

            if show_errorpage:
                return render_error(request)

            return redirect(
                redirect_url
                or reverse(
                    f"admin:{model._meta.app_label}_{model._meta.model_name}_changelist"
                )
            )

        return wrap

    return decorator


def require_all_kmuhelper_perms(
    permissions_required=[], login_url=None, raise_exception=True
):
    """
    Decorator for views that checks whether a user has ALL of the given kmuhelper
    permission enabled, redirecting to the log-in page if necessary.
    If the raise_exception parameter is given the PermissionDenied exception
    is raised.
    """

    def check_perms(user):
        if isinstance(permissions_required, str):
            perms = [permissions_required]
        else:
            perms = permissions_required

        perms = [f"kmuhelper.{perm}" if not "." in perm else perm for perm in perms]

        # First check if the user has the permission (even anon users)
        if user.has_perms(perms):
            return True
        # In case the 403 handler should be called raise the exception
        if raise_exception:
            raise PermissionDenied
        # As the last resort, show the login form
        return False

    return user_passes_test(check_perms, login_url=login_url)


def require_any_kmuhelper_perms(permissions=[], login_url=None, raise_exception=True):
    """
    Decorator for views that checks whether a user has ANY (of the given) kmuhelper
    permission enabled, redirecting to the log-in page if necessary.
    If the raise_exception parameter is given the PermissionDenied exception
    is raised.
    """

    def check_perms(user):
        if isinstance(permissions, str):
            perms = [permissions]
        else:
            perms = permissions

        # First check if the user has any kmuhelper permission
        if user.has_module_perms("kmuhelper"):
            # If no permissions are given, the user has access
            if not perms:
                return True

            # Check if the user has any of the given permissions
            for perm in perms:
                if user.has_perm(f"kmuhelper.{perm}" if not "." in perm else perm):
                    return True

        # In case the 403 handler should be called raise the exception
        if raise_exception:
            raise PermissionDenied
        # As the last resort, show the login form
        return False

    return user_passes_test(check_perms, login_url=login_url)


def require_kmuhelper_module_perms(module_name, login_url=None, raise_exception=True):
    """
    Decorator for views that checks whether a user has permissions for a kmuhelper
    module, redirecting to the log-in page if necessary.
    If the raise_exception parameter is given the PermissionDenied exception
    is raised.
    """

    def check_perms(user):
        # First check if the user has the permission (even anon users)
        if config.user_has_module_permission(user, module_name):
            return True

        # In case the 403 handler should be called raise the exception
        if raise_exception:
            raise PermissionDenied
        # As the last resort, show the login form
        return False

    return user_passes_test(check_perms, login_url=login_url)



from django.utils.translation import gettext

import kmuhelper.modules.main.models as models
from kmuhelper import constants
from kmuhelper.modules.integrations.woocommerce.api import products, customers
from kmuhelper.modules.integrations.woocommerce.api._base import WC_BaseObjectAPI
from kmuhelper.utils import runden

_ = gettext


class WCOrdersAPI(WC_BaseObjectAPI):
    LOG_PREFIX = "[deep_pink4][KMUHelper WooCommerce Orders][/] -"

    MODEL = models.Order
    WC_OBJ_DOES_NOT_EXIST_CODE = "woocommerce_rest_shop_order_invalid_id"
    WC_API_BASEURL = "orders"

    def update_object_from_data(self, db_obj, wc_obj: dict):
        db_obj.woocommerce_deleted = False

        if wc_obj["date_paid"]:
            db_obj.is_paid = True
        db_obj.status = wc_obj["status"]
        db_obj.customer_note = wc_obj["customer_note"]

        db_obj.addr_billing_first_name = wc_obj["billing"]["first_name"]
        db_obj.addr_billing_last_name = wc_obj["billing"]["last_name"]
        db_obj.addr_billing_company = wc_obj["billing"]["company"]
        db_obj.addr_billing_address_1 = wc_obj["billing"]["address_1"]
        db_obj.addr_billing_address_2 = wc_obj["billing"]["address_2"]
        db_obj.addr_billing_city = wc_obj["billing"]["city"]
        db_obj.addr_billing_state = wc_obj["billing"]["state"]
        db_obj.addr_billing_postcode = wc_obj["billing"]["postcode"]
        db_obj.addr_billing_country = wc_obj["billing"]["country"]
        db_obj.addr_billing_email = wc_obj["billing"]["email"]
        db_obj.addr_billing_phone = wc_obj["billing"]["phone"]

        db_obj.addr_shipping_first_name = wc_obj["shipping"]["first_name"]
        db_obj.addr_shipping_last_name = wc_obj["shipping"]["last_name"]
        db_obj.addr_shipping_company = wc_obj["shipping"]["company"]
        db_obj.addr_shipping_address_1 = wc_obj["shipping"]["address_1"]
        db_obj.addr_shipping_address_2 = wc_obj["shipping"]["address_2"]
        db_obj.addr_shipping_city = wc_obj["shipping"]["city"]
        db_obj.addr_shipping_state = wc_obj["shipping"]["state"]
        db_obj.addr_shipping_postcode = wc_obj["shipping"]["postcode"]
        db_obj.addr_shipping_country = wc_obj["shipping"]["country"]
        db_obj.save()
        self.log("Order updated: ", str(db_obj))

    def create_object_from_data(self, wc_obj: dict, sendstockwarning=True):
        db_obj = models.Order.objects.create(
            woocommerceid=wc_obj["id"],
            status=wc_obj["status"],
            is_shipped=(True if wc_obj["status"] == "completed" else False),
            is_removed_from_stock=(True if wc_obj["status"] == "completed" else False),
            payment_method=wc_obj["payment_method"],
            is_paid=(True if wc_obj["date_paid"] else False),
            customer_note=wc_obj["customer_note"],
            order_key=wc_obj["order_key"],
            addr_billing_first_name=wc_obj["billing"]["first_name"],
            addr_billing_last_name=wc_obj["billing"]["last_name"],
            addr_billing_company=wc_obj["billing"]["company"],
            addr_billing_address_1=wc_obj["billing"]["address_1"],
            addr_billing_address_2=wc_obj["billing"]["address_2"],
            addr_billing_city=wc_obj["billing"]["city"],
            addr_billing_state=wc_obj["billing"]["state"],
            addr_billing_postcode=wc_obj["billing"]["postcode"],
            addr_billing_country=wc_obj["billing"]["country"],
            addr_billing_email=wc_obj["billing"]["email"],
            addr_billing_phone=wc_obj["billing"]["phone"],
            addr_shipping_first_name=wc_obj["shipping"]["first_name"],
            addr_shipping_last_name=wc_obj["shipping"]["last_name"],
            addr_shipping_company=wc_obj["shipping"]["company"],
            addr_shipping_address_1=wc_obj["shipping"]["address_1"],
            addr_shipping_address_2=wc_obj["shipping"]["address_2"],
            addr_shipping_city=wc_obj["shipping"]["city"],
            addr_shipping_state=wc_obj["shipping"]["state"],
            addr_shipping_postcode=wc_obj["shipping"]["postcode"],
            addr_shipping_country=wc_obj["shipping"]["country"],
        )
        db_obj.date = wc_obj["date_created_gmt"] + "+00:00"
        if wc_obj["customer_id"]:
            customer, created = models.Customer.objects.get_or_create(
                woocommerceid=int(wc_obj["customer_id"])
            )
            if created:
                customers.WCCustomersAPI(self.wcapi).update_object_from_api(customer)

            db_obj.customer = customer
            db_obj.addr_shipping_email = customer.addr_shipping_email
            db_obj.addr_shipping_phone = customer.addr_shipping_phone

        for item in wc_obj["line_items"]:
            # Use variation id if available and != 0
            if "variation_id" in item and item["variation_id"]:
                product_id = item["variation_id"]
            else:
                product_id = item["product_id"]

            quantity = int(item["quantity"])
            product_price = runden(float(item["subtotal"]) / quantity)
            # subtotal/quantity is used instead of price to exclude discounts => handled under "coupon_lines"

            if product_id != 0:
                product, created = models.Product.objects.get_or_create(
                    woocommerceid=int(product_id)
                )
                if created:
                    products.WCProductsAPI(self.wcapi).update_object_from_api(product)

                db_obj.products.through.objects.create(
                    order=db_obj,
                    linked_product=product,
                    quantity=quantity,
                    product_price=product_price,
                )
            else:
                self.log("Product ID in a line item was 0!")
                db_obj.products.through.objects.create(
                    order=db_obj,
                    linked_product=None,
                    quantity=quantity,
                    product_price=product_price,
                    name=item["name"],
                    article_number=str(item["sku"]),
                )

        for item in wc_obj["shipping_lines"]:
            db_obj.fees.through.objects.create(
                order=db_obj,
                name=item["method_title"],
                price=float(item["total"]),
                vat_rate=(
                    constants.VAT_RATE_DEFAULT if float(item["total_tax"]) > 0 else 0
                ),
            )

        for item in wc_obj["fee_lines"]:
            db_obj.fees.through.objects.create(
                order=db_obj,
                name=item["name"],
                price=float(item["total"]),
                vat_rate=(
                    constants.VAT_RATE_DEFAULT if float(item["total_tax"]) > 0 else 0
                ),
            )

        for item in wc_obj["coupon_lines"]:
            db_obj.fees.through.objects.create(
                order=db_obj,
                name=_("Coupon: %(code)s") % {"code": item["code"]},
                price=-float(item["discount"]),
                vat_rate=(
                    constants.VAT_RATE_DEFAULT if float(item["discount_tax"]) > 0 else 0
                ),
            )

        db_obj.save()
        db_obj.second_save()
        self.log("Order created:", str(db_obj))

        if sendstockwarning:
            db_obj.email_stock_warning()
        return db_obj



from django.urls import path, re_path

from kmuhelper.modules.api import views

#######################

urlpatterns = [
    path("versions", views.versions, name="api-versions"),
    path("orders/unpaid", views.orders_unpaid, name="api-orders-unpaid"),
    path(
        "orders/<int:object_id>/set-paid",
        views.orders_set_paid,
        name="api-orders-set-paid",
    ),
    re_path(".*", views.not_found, name="api-not-found"),
]



from django.utils.html import strip_tags


def preparestring(string):
    """Prepare a HTML string for import"""
    return (
        strip_tags(string.replace("</p>", " ").replace("</strong>", " "))
        .replace("&#8211;", "-")
        .replace("&#8211;", " ")
        .replace("&#215;", "x")
        .replace("&#8220;", '"')
        .replace("&#8221;", '"')
        .replace("&nbsp;", " ")
        .replace("&lt;", "<")
        .replace("&gt;", ">")
        .replace("&amp;", "&")
        .replace("&quot;", '"')
        .replace("&#8222;", '"')
        .replace("  ", " ")
    )



import json
from random import randint
from urllib.parse import urlencode

from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.core.mail import mail_admins
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.urls import reverse, reverse_lazy
from django.utils.html import format_html
from django.utils.translation import gettext_lazy, gettext
from django.views.decorators.csrf import csrf_exempt
from rich import print

from kmuhelper import settings
from kmuhelper.decorators import (
    require_object,
    require_all_kmuhelper_perms,
    require_any_kmuhelper_perms,
    confirm_action,
)
from kmuhelper.modules.integrations.woocommerce.api import (
    WCGeneralAPI,
    WCCustomersAPI,
    WCOrdersAPI,
    WCProductsAPI,
    WCProductCategoriesAPI,
)
from kmuhelper.modules.integrations.woocommerce.forms import WooCommerceSettingsForm
from kmuhelper.modules.integrations.woocommerce.utils import (
    is_connected,
    base64_hmac_sha256,
    random_secret,
    test_wc_url,
)
from kmuhelper.modules.main.models import Product, Customer, ProductCategory, Order
from kmuhelper.utils import render_error

_ = gettext_lazy


def log(string, *args):
    print("[deep_pink4][KMUHelper][/] -", string, *args)


NOT_CONNECTED_ERRMSG = _("WooCommerce wurde scheinbar nicht richtig verbunden!")

#####


@csrf_exempt
def wc_auth_key(request):
    """Endpoint for receiving the keys from WooCommerce"""

    try:
        data = json.loads(request.body)
        storeurl = request.headers.get("user-agent").split(";")[1].lstrip()
        savedurl = settings.get_db_setting("wc-url")

        storedomain = storeurl.lstrip("https://").lstrip("http://").split("/")[0]
        saveddomain = savedurl.lstrip("https://").lstrip("http://").split("/")[0]

        if saveddomain == storedomain:
            settings.set_secret_db_setting("wc-url", storeurl)
            settings.set_secret_db_setting("wc-consumer_key", data["consumer_key"])
            settings.set_secret_db_setting(
                "wc-consumer_secret", data["consumer_secret"]
            )

            return JsonResponse({"success": True}, status=200)

        return JsonResponse({"success": False, "reason": "Unknown URL!"}, status=401)
    except KeyError:
        return JsonResponse({"success": False, "reason": "Useless data!"}, status=400)


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_any_kmuhelper_perms()
def wc_auth_end(request):
    if request.GET.get("success") == "1":
        messages.success(request, gettext("WooCommerce erfolgreich verbunden!"))
    else:
        messages.error(request, gettext("WooCommerce konnte nicht verbunden werden!"))
    return redirect(reverse("kmuhelper:wc-settings"))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["change_setting"])
def wc_auth_start(request):
    shopurl = settings.get_db_setting("wc-url", None)

    if not shopurl or not shopurl.startswith("http"):
        messages.error(
            request,
            gettext("Please enter a valid WooCommerce URL, beginning with 'https://'!"),
        )
        return redirect(reverse("kmuhelper:wc-settings"))

    if not request.is_secure():
        # WooCommerce only accepts callback_urls via SSL
        messages.error(request, gettext("Setting up WooCommerce is only possible when connected via HTTPS!"))
        return redirect(reverse("kmuhelper:wc-settings"))

    params = {
        "app_name": "KMUHelper",
        "scope": "read",
        "user_id": randint(100000, 999999),
        "return_url": request.build_absolute_uri(reverse("kmuhelper:wc-auth-end")),
        "callback_url": request.build_absolute_uri(reverse("kmuhelper:wc-auth-key")),
    }
    query_string = urlencode(params)

    url = "%s%s?%s" % (shopurl, "/wc-auth/v1/authorize", query_string)
    return redirect(url)


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["change_setting"])
@confirm_action(_("Delete stored API credentials"))
def wc_auth_clear(request):
    wc_url = settings.get_secret_db_setting("wc-url")
    wc_consumer_key = settings.get_secret_db_setting("wc-consumer_key")

    if not wc_url or not wc_consumer_key:
        messages.warning(request, gettext("WooCommerce was not set up."))
    else:
        api_keys_url = (
            wc_url + "/wp-admin/admin.php?page=wc-settings&tab=advanced&section=keys"
        )

        settings.set_secret_db_setting("wc-url", "")
        settings.set_secret_db_setting("wc-consumer_key", "")
        settings.set_secret_db_setting("wc-consumer_secret", "")

        messages.success(
            request,
            format_html(
                '{} <a target="_blank" href="{}">{}</a> (Consumer key: {})',
                gettext("Credentials deleted! Please revoke your API key manually:"),
                api_keys_url,
                gettext("click here"),
                wc_consumer_key,
            ),
        )
    return redirect(reverse("kmuhelper:wc-settings"))


# Note: Change permission is required because the view redirects to the settings page
@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["change_setting"])
def wc_system_status(request):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        status = WCGeneralAPI().get_system_status()

        if status:
            messages.success(request, _("WooCommerce is connected and works!"))
        else:
            messages.error(request, _("WooCommerce doesn't seem to work!"))
    return redirect(reverse("kmuhelper:wc-settings"))


# Note: Change permission is required because the view redirects to the settings page
@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["change_setting"])
def wc_webhooks_status(request):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        current_host = "https://" + request.get_host()

        if current_host.endswith("localhost"):
            messages.error(
                request, gettext("Cannot check webhooks while on localhost!")
            )
            return redirect(reverse("kmuhelper:wc-settings"))

        delivery_url = current_host + reverse("kmuhelper:wc-webhooks")
        success, topics = WCGeneralAPI().get_enabled_webhooks_topics(delivery_url)

        if success:
            if topics:
                messages.success(
                    request,
                    gettext("%(count)d topic(s) have been set up correctly: %(topics)s")
                    % {"count": len(topics), "topics": ",".join(topics)},
                )
            else:
                messages.warning(request, _("No correctly setup webhooks detected!"))
        else:
            messages.error(request, _("Checking webhook status failed!"))
    return redirect(reverse("kmuhelper:wc-settings"))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["add_product"])
def wc_import_products(request):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        WCProductsAPI().import_all_objects_from_api(request=request)
    return redirect(reverse("admin:kmuhelper_product_changelist"))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["add_customer"])
def wc_import_customers(request):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        WCCustomersAPI().import_all_objects_from_api(request=request)
    return redirect(reverse("admin:kmuhelper_customer_changelist"))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["add_productcategory"])
def wc_import_categories(request):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        WCProductCategoriesAPI().import_all_objects_from_api(request=request)
    return redirect(reverse("admin:kmuhelper_productcategory_changelist"))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["add_order"])
def wc_import_orders(request):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        WCOrdersAPI().import_all_objects_from_api(request=request)
    return redirect(reverse("admin:kmuhelper_order_changelist"))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["change_product"])
@require_object(Product)
def wc_update_product(request, obj):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        try:
            WCProductsAPI().update_object_from_api(obj)
            messages.success(request, gettext("Product '%s' updated!") % str(obj))
        except Exception as e:
            messages.error(
                request,
                gettext("Product '%(name)s' update failed! %(error_msg)s")
                % {"name": str(obj), "error_msg": str(e)},
            )
    return redirect(reverse("admin:kmuhelper_product_change", args=[obj.pk]))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["change_customer"])
@require_object(Customer)
def wc_update_customer(request, obj):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        try:
            WCCustomersAPI().update_object_from_api(obj)
            messages.success(request, gettext("Customer '%s' updated!") % str(obj))
        except Exception as e:
            messages.error(
                request,
                gettext("Customer '%(name)s' update failed! %(error_msg)s")
                % {"name": str(obj), "error_msg": str(e)},
            )
    return redirect(reverse("admin:kmuhelper_customer_change", args=[obj.pk]))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["change_productcategory"])
@require_object(ProductCategory)
def wc_update_category(request, obj):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        try:
            WCProductCategoriesAPI().update_object_from_api(obj)
            messages.success(
                request, gettext("Product category '%s' updated!") % str(obj)
            )
        except Exception as e:
            messages.error(
                request,
                gettext("Product category '%(name)s' update failed! %(error_msg)s")
                % {"name": str(obj), "error_msg": str(e)},
            )
    return redirect(reverse("admin:kmuhelper_productcategory_change", args=[obj.pk]))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["change_order"])
@require_object(Order)
def wc_update_order(request, obj):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        try:
            WCOrdersAPI().update_object_from_api(obj)
            messages.success(request, gettext("Order '%s' updated!") % str(obj))
        except Exception as e:
            messages.error(
                request,
                gettext("Order '%(name)s' update failed! %(error_msg)s")
                % {"name": str(obj), "error_msg": str(e)},
            )
    return redirect(reverse("admin:kmuhelper_order_change", args=[obj.pk]))


@csrf_exempt
def wc_webhooks(request):
    """Endpoint for WooCommerce webhooks"""

    # 1. Check request method

    if request.method != "POST":
        messages.warning(
            request,
            gettext(
                "This endpoint is only available via POST and not meant to be used by humans!"
            ),
        )
        return render_error(request, status=405)

    # 2. Check if WooCommerce is connected

    if not is_connected():
        return JsonResponse(
            {
                "accepted": False,
                "reason": "Webhooks are currently disabled.",
            },
            status=418,
        )

    # 3. Check request source

    if not (
        "x-wc-webhook-topic" in request.headers
        and "x-wc-webhook-source" in request.headers
    ):
        log(
            "[orange_red1]WooCommerce Webhook accepted but ignored (no usable headers)![/] "
            + "(Might be a test request from WooCommerce)"
        )
        return JsonResponse(
            {
                "accepted": True,
                "reason": "Request was accepted but ignored because it doesn't contain any usable info!",
            },
            status=200,
        )

    stored_url = (
        settings.get_secret_db_setting("wc-url")
        .lstrip("https://")
        .lstrip("http://")
        .split("/")[0]
    )
    received_url = (
        request.headers["x-wc-webhook-source"]
        .lstrip("https://")
        .lstrip("http://")
        .split("/")[0]
    )

    if not received_url == stored_url:
        log(
            "[orange_red1]WooCommerce Webhook rejected (unexpected domain)![/] "
            + "- Expected:",
            stored_url,
            "- Received:",
            received_url,
        )
        return JsonResponse(
            {
                "accepted": False,
                "reason": "Unknown domain!",
            },
            status=403,
        )

    # 4. Check request signature

    secret = settings.get_db_setting("wc-webhook-secret")

    if secret:
        if not ("x-wc-webhook-signature" in request.headers):
            log("[orange_red1]WooCommerce Webhook rejected (no signature)![/]")
            return JsonResponse(
                {
                    "accepted": False,
                    "reason": "No signature provided!",
                },
                status=403,
            )

        received_signature = bytes(
            request.headers["x-wc-webhook-signature"], encoding="utf-8"
        )
        expected_signature = base64_hmac_sha256(str(secret).encode(), request.body)

        if received_signature == expected_signature:
            log("[green]WooCommerce Webhook signature check successful![/]")
        else:
            log(
                "[orange_red1]WooCommerce Webhook signature check failed![/] "
                + "- Expected:",
                expected_signature,
                "- Received:",
                received_signature,
            )
            return JsonResponse(
                {
                    "accepted": False,
                    "reason": "Signature mismatch!",
                },
                status=403,
            )

    else:
        log(
            "[orange_red1]Skipped WooCommerce Webhook signature check (no secret available)![/]"
        )

    # Do stuff

    log("WooCommerce Webhook is being processed...")

    delivery_id = request.headers["x-wc-webhook-delivery-id"]
    topic = request.headers["x-wc-webhook-topic"]
    wc_obj = json.loads(request.body)
    wc_obj_id = wc_obj.get("id")

    log("Delivery ID: ", delivery_id)
    log("Topic: ", topic)
    log("Object ID: ", wc_obj_id)

    if not wc_obj_id:
        log("[orange_red1]Object ID was not provided - reporting to admins...[/]")
        mail_admins(
            "ERROR: WooCommerce Webhook without object ID received!",
            f"Delivery ID: {delivery_id}\nTopic: {topic}\nObject: {str(wc_obj)}",
        )
        return JsonResponse(
            {"accepted": False, "reason": "Object ID was not provided!"}, status=400
        )

    match topic:
        case "product.updated" | "product.created" | "product.restored":
            if Product.objects.filter(woocommerceid=wc_obj_id).exists():
                product = Product.objects.get(woocommerceid=wc_obj_id)
                WCProductsAPI().update_object_from_data(
                    product, wc_obj, is_create_event=(topic == "product.created")
                )
            else:
                WCProductsAPI().create_object_from_data(wc_obj)
        case "product.deleted":
            if Product.objects.filter(woocommerceid=wc_obj_id).exists():
                product = Product.objects.get(woocommerceid=wc_obj_id)
                WCProductsAPI().delete_object_from_data(product, wc_obj)
        case "customer.updated" | "customer.created" | "customer.restored":
            if Customer.objects.filter(woocommerceid=wc_obj_id).exists():
                customer = Customer.objects.get(woocommerceid=wc_obj_id)
                customer.woocommerce_deleted = False
                WCCustomersAPI().update_object_from_data(customer, wc_obj)
            else:
                WCCustomersAPI().create_object_from_data(wc_obj)
        case "customer.deleted":
            if Customer.objects.filter(woocommerceid=wc_obj_id).exists():
                customer = Customer.objects.get(woocommerceid=wc_obj_id)
                WCCustomersAPI().delete_object_from_data(customer, wc_obj)
        case "order.updated" | "order.created" | "order.restored":
            if Order.objects.filter(woocommerceid=wc_obj_id).exists():
                order = Order.objects.get(woocommerceid=wc_obj_id)
                order.woocommerce_deleted = False
                WCOrdersAPI().update_object_from_data(order, wc_obj)
            else:
                WCOrdersAPI().create_object_from_data(wc_obj)
        case "order.deleted":
            if Order.objects.filter(woocommerceid=wc_obj_id).exists():
                order = Order.objects.get(woocommerceid=wc_obj_id)
                WCOrdersAPI().delete_object_from_data(order, wc_obj)
        case _:
            log(f"[orange_red1]Unknown topic: '{topic}' - reporting to admins...")
            mail_admins(
                "ERROR: WooCommerce Webhook with unknown topic received!",
                f"Delivery ID: {delivery_id}\nTopic: {topic}\nObject: {str(wc_obj)}",
            )
            return JsonResponse(
                {"accepted": False, "message": "Topic not supported!"}, status=400
            )

    return JsonResponse(
        {"accepted": True, "message": "Successfully processed!"}, status=200
    )


# Settings


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["change_setting"])
def wc_settings(request):
    if request.method == "POST":
        form = WooCommerceSettingsForm(request.POST)
        if form.is_valid():
            form.save_settings()
            messages.success(request, gettext("Einstellungen gespeichert!"))
            # Redirect to prevent resending the form on reload
            return redirect("kmuhelper:wc-settings")
    else:
        form = WooCommerceSettingsForm()

    secreturl = settings.get_secret_db_setting("wc-url", None)
    url = settings.get_db_setting("wc-url", None)

    return render(
        request,
        "kmuhelper/integrations/woocommerce/settings.html",
        {
            "form": form,
            "has_permission": True,
            "random_secret": random_secret(),
            "kmuhelper_url": "https://" + request.get_host(),
            "wc_url": secreturl,
            "is_connected": is_connected(),
            "is_url_valid": test_wc_url(url),
        },
    )



# Test settings for django-kmuhelper

DEBUG = True
SECRET_KEY = "fa^c^jx6$epw$%#b3bc)=cn_dp@misn+5l-cogxw$w3lsdl3r="
DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3"}}
ROOT_URLCONF = "kmuhelper_tests.urls"
MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]
INSTALLED_APPS = (
    "kmuhelper_tests",
    "kmuhelper",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
)
TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]
STATIC_URL = "/static/"

# KMUHelper specific settings

KMUHELPER_DOMAIN = ""
KMUHELPER_LOG_EMAIL = ""
KMUHELPER_EMAILS_DEFAULT_CONTEXT = {
    "precontent": "",
    "postcontent": "",
    "header_title": "KMUHelper TESTS",
}



"""
Tests custom admin URLs to check for unexpected 500 errors caused by mistakes in views or templates.
"""

from django.conf import settings
from django.contrib.auth.models import User
from django.test import Client, TestCase

from kmuhelper.modules.emails.models import EMail
from kmuhelper.modules.main.models import ContactPerson, PaymentReceiver, Order, Supply


def _create_test_order():
    return Order.objects.create(
        contact_person=ContactPerson.objects.create(
            name="X X", phone="x", email="x@x"
        ),
        payment_receiver=PaymentReceiver.objects.create(),
    )


class CustomAdminURLTests(TestCase):
    def setUp(self):
        # Every test needs a client.
        self.client = Client()
        self.client.cookies.load({settings.LANGUAGE_COOKIE_NAME: "de"})
        self.user = User.objects.create_user(
            username="testuser", password="<PASSWORD>", is_staff=True, is_superuser=True
        )
        self.client.force_login(self.user)

    # Utils

    def _require_successful_get(self, url, code=200):
        """Checks that a GET request to a given URL is successful and returns the request object."""
        response = self.client.get(url)
        print(f"Checking GET to {url}")
        self.assertEqual(response.status_code, code, msg=f"Couldn't GET {url}, code {response.status_code}")
        return response

    def _require_successful_post(self, url, data=None, code=200):
        """Checks that a POST request to a given URL is successful and returns the request object."""
        response = self.client.post(url, data)
        print(f"Checking POST to {url}")
        self.assertEqual(response.status_code, code, msg=f"Couldn't POST {url}, code {response.status_code}")
        return response

    # Custom order admin URLs

    def test_admin_order_duplicate(self):
        obj = _create_test_order()
        self._require_successful_post(f"/admin/kmuhelper/order/{obj.pk}/duplicate/", code=302)

        count = Order.objects.count()
        self.assertEqual(count, 2, msg="Did not create any order!")

    def test_admin_order_return(self):
        obj = _create_test_order()
        self._require_successful_post(f"/admin/kmuhelper/order/{obj.pk}/return/", code=302)

        count = Supply.objects.count()
        self.assertEqual(count, 1, msg="Did not create any supply!")

    def test_admin_order_email_invoice(self):
        obj = _create_test_order()
        self._require_successful_post(f"/admin/kmuhelper/order/{obj.pk}/email/invoice/", code=302)

        count = EMail.objects.count()
        self.assertEqual(count, 1, msg="Did not create any email!")

    def test_admin_order_email_shipped(self):
        obj = _create_test_order()
        self._require_successful_post(f"/admin/kmuhelper/order/{obj.pk}/email/shipped/", code=302)

        count = EMail.objects.count()
        self.assertEqual(count, 1, msg="Did not create any email!")

    def test_admin_order_pdf(self):
        obj = _create_test_order()
        response = self._require_successful_get(f"/admin/kmuhelper/order/{obj.pk}/pdf/")
        self.assertEqual(response.headers.get('Content-Type'), 'application/pdf', msg="Returned wrong content type!")

    def test_admin_order_pdf_form(self):
        obj = _create_test_order()
        self._require_successful_get(f"/admin/kmuhelper/order/{obj.pk}/pdf/form/", code=302)



# Generated by Django 5.2 on 2025-04-08 21:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kmuhelper', '0116_add_woocommerceid_indexes'),
    ]

    operations = [
        migrations.AlterField(
            model_name='order',
            name='fees',
            field=models.ManyToManyField(through='kmuhelper.OrderFee', through_fields=('order', 'linked_fee'), to='kmuhelper.fee'),
        ),
        migrations.AlterField(
            model_name='order',
            name='products',
            field=models.ManyToManyField(through='kmuhelper.OrderItem', through_fields=('order', 'linked_product'), to='kmuhelper.product'),
        ),
        migrations.AlterField(
            model_name='product',
            name='categories',
            field=models.ManyToManyField(related_name='products', through='kmuhelper.ProductProductCategoryConnection', through_fields=('product', 'category'), to='kmuhelper.productcategory', verbose_name='Kategorien'),
        ),
        migrations.AlterField(
            model_name='supply',
            name='products',
            field=models.ManyToManyField(through='kmuhelper.SupplyItem', through_fields=('supply', 'product'), to='kmuhelper.product'),
        ),
    ]



from django.contrib import admin
from django.db import models
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from kmuhelper import settings

WC_STATE_DESCRIPTION = format_html(
    '<abbr title="{}"><i class="fa-solid fa-cloud"></i></abbr>',
    _(
        "WooCommerce-Status (grün = verknüpft, rot = verknüpft, aber auf WooCommerce gelöscht, grau = nicht verknüpft)"
    ),
)


class WooCommerceModelMixin(models.Model):
    WOOCOMMERCE_URL_FORMAT = "{}/wp-admin/post.php?action=edit&post={}"

    woocommerceid = models.IntegerField(
        verbose_name=_("WooCommerce ID"),
        default=0,
    )

    woocommerce_deleted = models.BooleanField(
        verbose_name=_("Deleted in WooCommerce?"),
        default=False,
        blank=True,
        null=False,
    )

    def get_woocommerce_url(self):
        return self.WOOCOMMERCE_URL_FORMAT.format(
            settings.get_secret_db_setting("wc-url"), self.woocommerceid
        )

    # Display

    @admin.display(description="WooCommerce", ordering="woocommerceid")
    def display_woocommerce_id(self):
        if not self.woocommerceid:
            return None

        link = self.get_woocommerce_url()

        return format_html(
            '<a target="_blank" href="{}">#{}</a> {}',
            link,
            str(self.woocommerceid),
            _("(Deleted)") if self.woocommerce_deleted else "",
        )

    @admin.display(
        description=WC_STATE_DESCRIPTION,
        ordering="woocommerceid",
        boolean=True,
    )
    def display_woocommerce_state(self):
        if not self.woocommerceid:
            return None
        return not self.woocommerce_deleted

    class Meta:
        abstract = True

        indexes = [models.Index(fields=["woocommerceid"])]



import kmuhelper.modules.main.models
import kmuhelper.modules.app.models
import kmuhelper.modules.emails.models
import kmuhelper.modules.api.models
import kmuhelper.modules.settings.models

import kmuhelper.modules.integrations.paymentimport.models



"""
Tests for the order model and corresponding orderitem and orderfee models
"""

from django.test import TestCase

from kmuhelper.constants import VAT_RATE_DEFAULT
from kmuhelper.modules.main.models import (
    Order,
    OrderItem,
    OrderFee,
    ContactPerson,
    PaymentReceiver,
)
from kmuhelper.utils import runden


class OrderTest(TestCase):
    def setUp(self):
        self.contact_person = ContactPerson.objects.create(
            name="X X", phone="x", email="x@x"
        )
        self.payment_receiver = PaymentReceiver.objects.create()

    # Order total

    def test_order_total__no_vat(self):
        obj = Order.objects.create(
            contact_person=self.contact_person,
            payment_receiver=self.payment_receiver,
        )

        obj.products.through.objects.bulk_create(
            [
                OrderItem(product_price=10.5, quantity=1, vat_rate=0.0, order=obj),
                OrderItem(product_price=5.05, quantity=2, vat_rate=0.0, order=obj),
            ]
        )

        obj.fees.through.objects.bulk_create(
            [
                OrderFee(price=9.95, vat_rate=0.0, order=obj),
            ]
        )

        total_expected = 10.5 + (5.05 * 2) + 9.95
        total_calculated = obj.calc_total()

        self.assertAlmostEqual(
            total_expected, total_calculated, 2, "Order total does not match!"
        )

    def test_order_total__with_vat(self):
        obj = Order.objects.create(
            contact_person=self.contact_person,
            payment_receiver=self.payment_receiver,
        )

        obj.products.through.objects.bulk_create(
            [
                OrderItem(
                    product_price=10.5, quantity=1, vat_rate=VAT_RATE_DEFAULT, order=obj
                ),
                OrderItem(
                    product_price=5.05, quantity=2, vat_rate=VAT_RATE_DEFAULT, order=obj
                ),
            ]
        )

        obj.fees.through.objects.bulk_create(
            [
                OrderFee(price=9.95, vat_rate=VAT_RATE_DEFAULT, order=obj),
            ]
        )

        VAT_FACTOR = 1 + 0.01 * VAT_RATE_DEFAULT
        total_expected = runden(
            runden(10.5 * VAT_FACTOR)
            + runden(5.05 * 2 * VAT_FACTOR)
            + runden(9.95 * VAT_FACTOR)
        )
        total_calculated = obj.calc_total()

        self.assertEqual(total_expected, total_calculated)

    # Order item discount

    def test_order_item__discount(self):
        item = OrderItem(product_price=5.55, quantity=17, discount=10, vat_rate=0.0)

        price_expected = runden(5.55 * 17 * (1 - 0.1))
        price_calculated = item.calc_subtotal()

        self.assertEqual(price_expected, price_calculated)



from django.contrib import admin
from django.urls import reverse, NoReverseMatch
from django.utils.html import format_html
from django.utils.translation import gettext_lazy
from kmuhelper.modules.log.models import AdminLogEntry
from kmuhelper.overrides import CustomModelAdmin

_ = gettext_lazy


@admin.register(AdminLogEntry)
class AdminLogEntryAdmin(CustomModelAdmin):
    HIDDEN = True

    list_filter = [
        "action_flag",
        "action_time",
        ("content_type", admin.RelatedOnlyFieldListFilter),
        "user",
    ]
    list_display = [
        "action_time",
        "content_type_display",
        "object_id_display",
        "object_repr",
        "action_flag",
        "action_display",
        "user_display",
    ]

    search_fields = ["object_id", "object_repr", "change_message"]

    ordering = ["-action_time"]

    date_hierarchy = "action_time"

    @admin.display(description=_("Typ"), ordering="content_type")
    def content_type_display(self, obj):
        if obj.content_type is None:
            return None
        try:
            url = reverse(
                "admin:{}_{}_changelist".format(
                    obj.content_type.app_label, obj.content_type.model
                )
            )
            return format_html(
                '<a href="{}">{}</a>',
                url,
                obj.content_type.name,
            )
        except NoReverseMatch:
            return obj.content_type.name

    @admin.display(description=_("ID"), ordering="object_id")
    def object_id_display(self, obj):
        url = obj.get_admin_url()
        if url:
            return format_html(
                '<a href="{}">#{}</a>',
                obj.get_admin_url(),
                obj.object_id,
            )
        else:
            return f"#{obj.object_id}"

    @admin.display(description=_("Aktion"))
    def action_display(self, obj):
        return obj.get_change_message()

    @admin.display(description=_("Benutzer"), ordering="user_id")
    def user_display(self, obj):
        return format_html(
            '<a href="{}">{}</a>',
            reverse("admin:auth_user_change", args=(obj.user_id,)),
            str(obj.user),
        )

    # Permissions

    NO_CHANGE = True
    NO_ADD = True
    NO_DELETE = True


#

modeladmins = [
    (AdminLogEntry, AdminLogEntryAdmin),
]



from django.contrib import admin, messages
from django.db.models import Count
from django.urls import path, reverse
from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import gettext_lazy, ngettext

from kmuhelper import constants
from kmuhelper.modules.integrations.woocommerce.api import (
    WCCustomersAPI,
    WCOrdersAPI,
    WCProductsAPI,
    WCProductCategoriesAPI,
)
from kmuhelper.modules.integrations.woocommerce.filters import WooCommerceStateFilter
from kmuhelper.modules.integrations.woocommerce.utils import is_connected
from kmuhelper.modules.main import views
from kmuhelper.modules.main.filters import ProductTypeFilter
from kmuhelper.modules.main.models import (
    ContactPerson,
    Order,
    ProductCategory,
    Fee,
    Customer,
    Supplier,
    Supply,
    Note,
    Product,
    PaymentReceiver,
)
from kmuhelper.modules.pdfgeneration.order import views as pdf_order_views
from kmuhelper.overrides import (
    CustomModelAdmin,
    CustomTabularInline,
    CustomStackedInline,
)
from kmuhelper.utils import formatprice

_ = gettext_lazy

#######


@admin.register(ContactPerson)
class ContactPersonAdmin(CustomModelAdmin):
    fieldsets = [
        (None, {"fields": ["is_default"]}),
        (_("Name"), {"fields": ["name"]}),
        (_("Kontaktinformationen"), {"fields": ["phone", "email"]}),
    ]

    ordering = ("name",)

    list_display = ("pkfill", "name", "phone", "email", "is_default")
    search_fields = ["name", "phone", "email"]

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        if obj.is_default:
            ContactPerson.objects.filter(is_default=True).exclude(pk=obj.pk).update(
                is_default=False
            )


class OrderAdminOrderItemInline(CustomTabularInline):
    model = Order.products.through
    verbose_name = _("Bestellungsposten")
    verbose_name_plural = _("Bestellungsposten")
    extra = 0

    fields = (
        "display_product_link",
        "article_number",  # hidden by default
        "name",
        "note",
        "product_price",
        "quantity",
        "quantity_description",  # hidden by default
        "discount",
        "display_subtotal",
        "vat_rate",
    )

    readonly_fields = (
        "display_product_link",
        "display_subtotal",
    )

    def get_additional_readonly_fields(self, request, obj=None):
        fields = []
        if obj and (obj.is_shipped or obj.is_paid):
            fields += ["quantity"]
        if obj and obj.is_paid:
            fields += ["product_price", "vat_rate", "discount"]
        return fields

    # Display items

    @admin.display(description=_("Produkt"), ordering="linked_product")
    def display_product_link(self, obj):
        if obj.linked_product_id is None:
            return "-"

        link = reverse("admin:kmuhelper_product_change", args=(obj.linked_product_id,))

        return format_html(
            '<a href="{}">{}</a>', link, str(obj.linked_product_id).zfill(6)
        )

    @admin.display(description=_("Summe (exkl. MwSt)"))
    def display_subtotal(self, obj):
        return formatprice(obj.calc_subtotal()) + " CHF"

    # Permissions

    def has_add_permission(self, request, obj=None):
        return (
            False
            if (obj and (obj.is_paid or obj.is_shipped))
            else super().has_add_permission(request, obj)
        )

    def has_delete_permission(self, request, obj=None):
        return (
            False
            if (obj and (obj.is_shipped or obj.is_paid))
            else super().has_delete_permission(request, obj)
        )

    # Custom queryset

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related("linked_product")


class OrderAdminOrderItemInlineImport(CustomTabularInline):
    model = Order.products.through
    verbose_name = _("Bestellungsposten")
    verbose_name_plural = _("Bestellungsposten importieren")
    extra = 0

    autocomplete_fields = ("linked_product",)

    fields = (
        "linked_product",
        "note",
        "quantity",
        "discount",
    )

    # Permissions

    NO_CHANGE = True
    NO_DELETE = True
    NO_VIEW = True

    def has_add_permission(self, request, obj=None):
        return (
            False
            if (obj and (obj.is_shipped or obj.is_paid))
            else super().has_add_permission(request, obj)
        )


class OrderAdminOrderFeeInline(CustomTabularInline):
    model = Order.fees.through
    verbose_name = _("Bestellungskosten")
    verbose_name_plural = _("Bestellungskosten")
    extra = 0

    fields = (
        "display_fee_link",
        "name",
        "note",
        "price",
        "discount",
        "display_subtotal",
        "vat_rate",
    )

    readonly_fields = (
        "display_fee_link",
        "display_subtotal",
    )

    def get_additional_readonly_fields(self, request, obj=None):
        if obj and obj.is_paid:
            return ["price", "vat_rate", "discount"]
        return []

    # Display items

    @admin.display(description=_("Kosten"), ordering="linked_fee")
    def display_fee_link(self, obj):
        if obj.linked_fee_id is None:
            return "-"

        link = reverse("admin:kmuhelper_fee_change", args=(obj.linked_fee_id,))

        return format_html('<a href="{}">{}</a>', link, str(obj.linked_fee_id).zfill(4))

    @admin.display(description=_("Summe (exkl. MwSt)"))
    def display_subtotal(self, obj):
        return formatprice(obj.calc_subtotal()) + " CHF"

    # Permissions

    def has_add_permission(self, request, obj=None):
        return (
            False if (obj and obj.is_paid) else super().has_add_permission(request, obj)
        )

    def has_delete_permission(self, request, obj=None):
        return (
            False
            if (obj and obj.is_paid)
            else super().has_delete_permission(request, obj)
        )

    # Custom queryset

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related("linked_fee")


class OrderAdminOrderFeeInlineImport(CustomTabularInline):
    model = Order.fees.through
    verbose_name = _("Bestellungskosten")
    verbose_name_plural = _("Bestellungskosten importieren")
    extra = 0

    autocomplete_fields = ("linked_fee",)

    fields = (
        "linked_fee",
        "note",
        "discount",
    )

    # Permissions

    NO_CHANGE = True
    NO_DELETE = True
    NO_VIEW = True

    def has_add_permission(self, request, obj=None):
        return (
            False if (obj and obj.is_paid) else super().has_add_permission(request, obj)
        )


@admin.register(Order)
class OrderAdmin(CustomModelAdmin):
    list_display = [
        "pkfill",
        "date",
        "display_customer",
        "status",
        "payment_method",
        "display_is_shipped",
        "display_is_paid",
        "display_cached_sum",
        "linked_note_html",
    ]

    list_filter = (
        "status",
        "is_paid",
        "is_shipped",
        WooCommerceStateFilter,
        "payment_method",
        "payment_receiver",
        "contact_person",
    )
    search_fields = (
        [
            "id",
            "date",
            "linked_note__name",
            "linked_note__description",
            "customer_note",
            "tracking_number",
        ]
        + constants.ADDR_BILLING_FIELDS
        + constants.ADDR_SHIPPING_FIELDS
    )

    ordering = ("is_shipped", "is_paid", "-date")

    autocomplete_fields = (
        "customer",
        "payment_receiver",
        "contact_person",
    )

    save_on_top = True

    list_select_related = ["customer", "linked_note"]

    date_hierarchy = "date"

    def get_list_display(self, request):
        if is_connected():
            ls = self.list_display.copy()
            ls.insert(1, "display_woocommerce_state")
            return ls
        return self.list_display

    def get_fieldsets(self, request, obj=None):
        if obj:
            fieldsets = [
                (
                    _("Einstellungen"),
                    {"fields": ["payment_receiver", "contact_person"]},
                ),
                (_("Infos"), {"fields": ["name", "date", "status"]}),
                (_("Kunde"), {"fields": ["customer"]}),
                (
                    _("Lieferung"),
                    {"fields": [("shipped_on", "is_shipped"), "tracking_number"]},
                ),
                (
                    _("Bezahlungsoptionen"),
                    {
                        "fields": [
                            "payment_method",
                            "invoice_date",
                            "payment_conditions",
                            "payment_purpose",
                        ]
                    },
                ),
                (
                    _("Bezahlung"),
                    {
                        "fields": [
                            ("display_total_breakdown", "display_payment_conditions"),
                            ("paid_on", "is_paid"),
                        ]
                    },
                ),
                (
                    _("Notizen & Texte"),
                    {
                        "fields": ["customer_note", "linked_note_html"],
                        "classes": ["collapse start-open"],
                    },
                ),
                (
                    _("Rechnungsadresse"),
                    {
                        "fields": constants.ADDR_BILLING_FIELDS
                        if obj.is_paid
                        else constants.ADDR_BILLING_FIELDS_CATEGORIZED,
                        "classes": ["collapse start-open addr-billing-fieldset"],
                    },
                ),
                (
                    _("Lieferadresse"),
                    {
                        "fields": constants.ADDR_SHIPPING_FIELDS
                        if obj.is_shipped
                        else constants.ADDR_SHIPPING_FIELDS_CATEGORIZED,
                        "classes": ["collapse start-open addr-shipping-fieldset"],
                    },
                ),
            ]

            if obj.woocommerceid:
                fieldsets.insert(
                    1, (_("Verknüpfungen"), {"fields": ["display_woocommerce_id"]})
                )

            return fieldsets

        return [
            (_("Einstellungen"), {"fields": ["payment_receiver", "contact_person"]}),
            (_("Infos"), {"fields": ["status"]}),
            (_("Kunde"), {"fields": ["customer"]}),
            (
                _("Bezahlungsoptionen"),
                {
                    "fields": [
                        "payment_method",
                        "invoice_date",
                        "payment_conditions",
                        "payment_purpose",
                    ],
                    "classes": ["collapse start-open"],
                },
            ),
            (
                _("Notizen & Texte"),
                {"fields": ["customer_note"], "classes": ["collapse start-open"]},
            ),
        ]

    readonly_fields = (
        "linked_note_html",
        "name",
        "tracking_link",
        "display_total_breakdown",
        "display_payment_conditions",
        "display_woocommerce_id",
    )

    def get_additional_readonly_fields(self, request, obj=None):
        fields = []
        if obj:
            if obj.is_shipped:
                fields += [
                    "is_shipped"
                ] + constants.ADDR_SHIPPING_FIELDS_WITHOUT_CONTACT
            if obj.is_paid:
                fields += [
                    "is_paid",
                    "payment_method",
                    "invoice_date",
                    "payment_conditions",
                ] + constants.ADDR_BILLING_FIELDS_WITHOUT_CONTACT
            if obj.woocommerceid:
                fields += ["customer_note"]
        return fields

    def get_inlines(self, request, obj=None):
        inlines = [OrderAdminOrderItemInline]
        if request.user.has_perm("kmuhelper.view_product"):
            inlines += [OrderAdminOrderItemInlineImport]
        inlines += [OrderAdminOrderFeeInline]
        if request.user.has_perm("kmuhelper.view_fee"):
            inlines += [OrderAdminOrderFeeInlineImport]
        return inlines

    # Actions

    @admin.action(description=_("Als bezahlt markieren"), permissions=["change"])
    def mark_as_paid(self, request, queryset):
        successcount = 0
        errorcount = 0
        for order in queryset.all():
            if order.is_paid:
                errorcount += 1
            else:
                order.is_paid = True
                if order.is_shipped:
                    order.status = "completed"
                order.save()
                successcount += 1

        if successcount:
            messages.success(
                request,
                ngettext(
                    "%d Bestellung wurde als bezahlt markiert.",
                    "%d Bestellungen wurden als bezahlt markiert.",
                    successcount,
                )
                % successcount,
            )
        if errorcount:
            messages.warning(
                request,
                ngettext(
                    "%d Bestellung war bereits als bezahlt markiert.",
                    "%d Bestellungen waren bereits als bezahlt markiert.",
                    errorcount,
                )
                % errorcount,
            )

    @admin.action(
        description=_("Bestellungen von WooCommerce aktualisieren"),
        permissions=["change"],
    )
    def wc_update(self, request, queryset):
        WCOrdersAPI().bulk_update_objects_from_api(queryset.all(), request)

    actions = [mark_as_paid, wc_update]

    # Save

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)

        # stock warnings
        if obj:
            for product in obj.products.all():
                product.show_stock_warning(request)

    def save_related(self, request, form, formsets, change):
        super().save_related(request, form, formsets, change)

        form.instance.second_save()

    # Views

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name

        urls = super().get_urls()

        my_urls = [
            path(
                "<path:object_id>/pdf/",
                self.admin_site.admin_view(pdf_order_views.order_view_pdf),
                name="%s_%s_pdf" % info,
            ),
            path(
                "<path:object_id>/pdf/form",
                self.admin_site.admin_view(pdf_order_views.order_create_pdf_form),
                name="%s_%s_pdf_form" % info,
            ),
            path(
                "<path:object_id>/email/invoice/",
                self.admin_site.admin_view(views.create_order_email_invoice),
                name="%s_%s_email_invoice" % info,
            ),
            path(
                "<path:object_id>/email/shipped/",
                self.admin_site.admin_view(views.create_order_email_shipped),
                name="%s_%s_email_shipped" % info,
            ),
            path(
                "<path:object_id>/duplicate/",
                self.admin_site.admin_view(views.duplicate_order),
                name="%s_%s_duplicate" % info,
            ),
            path(
                "<path:object_id>/return/",
                self.admin_site.admin_view(views.copy_order_to_supply),
                name="%s_%s_copy_to_supply" % info,
            ),
        ]
        return my_urls + urls


@admin.register(Fee)
class FeeAdmin(CustomModelAdmin):
    list_display = ["pkfill", "clean_name", "price", "vat_rate"]

    search_fields = ("name", "price")

    ordering = (
        "price",
        "vat_rate",
    )

    fieldsets = [(None, {"fields": ("name", "price", "vat_rate")})]


class CustomerAdminOrderInline(CustomTabularInline):
    model = Order
    verbose_name = _("Bestellung")
    verbose_name_plural = _("Bestellungen")
    extra = 0

    show_change_link = True

    ordering = ("-date",)

    fields = (
        "pk",
        "date",
        "display_cached_sum",
        "is_shipped",
        "is_paid",
        "display_paid_after",
    )
    readonly_fields = ("pk", "display_paid_after", "display_cached_sum")

    # Custom queryset

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related("customer")

    # Permissions

    NO_CHANGE = True
    NO_ADD = True
    NO_DELETE = True


@admin.register(Customer)
class CustomerAdmin(CustomModelAdmin):
    list_display = [
        "pkfill",
        "company",
        "last_name",
        "first_name",
        "addr_billing_postcode",
        "addr_billing_city",
        "email",
        "avatar",
        "linked_note_html",
    ]

    ordering = ("addr_billing_postcode", "company", "last_name", "first_name")

    search_fields = (
        [
            "id",
            "last_name",
            "first_name",
            "company",
            "email",
            "username",
            "website",
            "linked_note__name",
            "linked_note__description",
        ]
        + constants.ADDR_BILLING_FIELDS
        + constants.ADDR_SHIPPING_FIELDS
    )

    readonly_fields = ["linked_note_html", "display_woocommerce_id"]

    list_filter = [WooCommerceStateFilter]

    list_select_related = ["linked_note"]

    autocomplete_fields = ["combine_with"]

    inlines = [CustomerAdminOrderInline]

    save_on_top = True

    def get_list_display(self, request):
        if is_connected():
            ls = self.list_display.copy()
            ls.insert(1, "display_woocommerce_state")
            return ls
        return self.list_display

    def get_fieldsets(self, request, obj=None):
        fieldsets = [
            (
                _("Infos"),
                {"fields": ["first_name", "last_name", "company", "email", "language"]},
            ),
            (
                _("Rechnungsadresse"),
                {
                    "fields": constants.ADDR_BILLING_FIELDS_CATEGORIZED,
                    "classes": ["addr-billing-fieldset"],
                },
            ),
            (
                _("Lieferadresse"),
                {
                    "fields": constants.ADDR_SHIPPING_FIELDS_CATEGORIZED,
                    "classes": ["addr-shipping-fieldset"],
                },
            ),
        ]

        if obj:
            if obj.woocommerceid:
                fieldsets.insert(
                    0, (_("Verknüpfungen"), {"fields": ["display_woocommerce_id"]})
                )
            return fieldsets + [
                (_("Diverses"), {"fields": ["website", "note", "linked_note_html"]}),
                (_("Erweitert"), {"fields": ["combine_with"], "classes": ["collapse"]}),
            ]

        return fieldsets + [(_("Diverses"), {"fields": ["website", "note"]})]

    # Actions

    @admin.action(
        description=_("Kunden von WooCommerce aktualisieren"), permissions=["change"]
    )
    def wc_update(self, request, queryset):
        WCCustomersAPI().bulk_update_objects_from_api(queryset.all(), request)

    actions = ["wc_update"]

    # Views

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name

        urls = super().get_urls()

        my_urls = [
            path(
                "<path:object_id>/email/registered/",
                self.admin_site.admin_view(views.create_customer_email_registered),
                name="%s_%s_email_registered" % info,
            ),
            path(
                "<path:object_id>/create-order",
                self.admin_site.admin_view(views.create_customer_order),
                name="%s_%s_create_order" % info,
            ),
        ]
        return my_urls + urls


@admin.register(Supplier)
class SupplierAdmin(CustomModelAdmin):
    fieldsets = [
        (_("Infos"), {"fields": ["abbreviation", "name"]}),
        (_("Firma"), {"fields": ["website", "phone", "email", "address"]}),
        (
            _("Ansprechpartner"),
            {
                "fields": [
                    "contact_person_name",
                    "contact_person_phone",
                    "contact_person_email",
                ],
                "classes": ["collapse", "start-open"],
            },
        ),
        (_("Notiz"), {"fields": ["note"], "classes": ["collapse", "start-open"]}),
    ]

    ordering = ("abbreviation",)

    list_display = ("pkfill", "abbreviation", "name", "note")
    search_fields = ["abbreviation", "name", "address", "note"]

    # Views

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name

        urls = super().get_urls()

        my_urls = [
            path(
                "<path:object_id>/assign/",
                self.admin_site.admin_view(views.supplier_assign),
                name="%s_%s_assign" % info,
            ),
        ]
        return my_urls + urls


class SupplyInlineSupplyItem(CustomTabularInline):
    model = Supply.products.through
    verbose_name = _("Produkt")
    verbose_name_plural = _("Produkte")
    extra = 0

    readonly_fields = ("product",)

    fields = (
        "product",
        "quantity",
    )

    # Permissions

    NO_ADD = True

    def has_change_permission(self, request, obj=None):
        return (
            False
            if obj and obj.is_added_to_stock
            else super().has_change_permission(request, obj)
        )

    def has_delete_permission(self, request, obj=None):
        return (
            False
            if obj and obj.is_added_to_stock
            else super().has_delete_permission(request, obj)
        )


class SupplyInlineProductsAdd(CustomTabularInline):
    model = Supply.products.through
    verbose_name = _("Produkt")
    verbose_name_plural = _("Produkte hinzufügen")
    extra = 0

    autocomplete_fields = ("product",)

    fields = (
        "product",
        "quantity",
    )

    # Permissions

    NO_VIEW = True
    NO_CHANGE = True
    NO_DELETE = True

    def has_add_permission(self, request, obj=None):
        if not (
            request.user.has_perm("kmuhelper.view_product")
            or request.user.has_perm("kmuhelper.change_product")
        ):
            return False
        return (
            False
            if obj and obj.is_added_to_stock
            else super().has_add_permission(request, obj)
        )


@admin.register(Supply)
class SupplyAdmin(CustomModelAdmin):
    list_display = (
        "pkfill",
        "name",
        "date",
        "total_quantity",
        "supplier",
        "is_added_to_stock",
        "linked_note_html",
    )
    list_filter = (
        "is_added_to_stock",
        "supplier",
    )

    search_fields = [
        "name",
        "date",
        "supplier__name",
        "supplier__abbreviation",
        "linked_note__name",
        "linked_note__description",
    ]

    readonly_fields = ["linked_note_html"]

    autocomplete_fields = ("supplier",)

    ordering = ("-date",)

    fieldsets = [
        (_("Infos"), {"fields": ["name", "linked_note_html"]}),
        (_("Lieferant"), {"fields": ["supplier"]}),
    ]

    inlines = [SupplyInlineSupplyItem, SupplyInlineProductsAdd]

    save_on_top = True

    list_select_related = (
        "supplier",
        "linked_note",
    )

    # Actions

    @admin.action(description=_("Lieferungen einlagern"), permissions=["change"])
    def add_to_stock(self, request, queryset):
        successcount = 0
        errorcount = 0
        for supply in queryset.all():
            if supply.add_to_stock():
                successcount += 1
            else:
                errorcount += 1

        if successcount:
            messages.success(
                request,
                ngettext(
                    "%d Lieferung wurde als eingelagert markiert.",
                    "%d Lieferungen wurden als eingelagert markiert.",
                    successcount,
                )
                % successcount,
            )
        if errorcount:
            messages.error(
                request,
                ngettext(
                    "%d Lieferung konnte nicht eingelagert werden.",
                    "%d Lieferungen konnten nicht eingelagert werden.",
                    errorcount,
                )
                % errorcount,
            )

    actions = ["add_to_stock"]

    # Views

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name

        urls = super().get_urls()

        my_urls = [
            path(
                "<path:object_id>/add_to_stock/",
                self.admin_site.admin_view(views.supply_add_to_stock),
                name="%s_%s_add_to_stock" % info,
            ),
        ]
        return my_urls + urls


@admin.register(Note)
class NoteAdmin(CustomModelAdmin):
    list_display = ["pkfill", "name", "description", "priority", "done", "created_at"]
    list_filter = ["done", "priority"]

    readonly_fields = ["links"]

    ordering = ["done", "-priority", "created_at"]

    search_fields = ("name", "description")

    def get_fieldsets(self, request, obj=None):
        if obj:
            return [
                (_("Infos"), {"fields": ["name", "description"]}),
                (_("Daten"), {"fields": ["done", "priority"]}),
                (_("Verknüpfungen"), {"fields": ["links"]}),
            ]

        return [
            (_("Infos"), {"fields": ["name", "description"]}),
            (_("Daten"), {"fields": ["done", "priority"]}),
        ]

    def get_form(self, request, obj=None, change=False, **kwargs):
        form = super().get_form(request, obj, change, **kwargs)
        if obj is None:
            form.base_fields["description"].initial = ""
            if "from_order" in request.GET:
                pk = request.GET.get("from_order")
                t = _("Bestellung #%s") % pk
                form.base_fields["name"].initial = t
                form.base_fields["description"].initial += f"\n\n[{t}]"
            if "from_product" in request.GET:
                pk = request.GET.get("from_product")
                t = _("Produkt #%s") % pk
                form.base_fields["name"].initial = t
                form.base_fields["description"].initial += f"\n\n[{t}]"
            if "from_customer" in request.GET:
                pk = request.GET.get("from_customer")
                t = _("Kunde #%s") % pk
                form.base_fields["name"].initial = t
                form.base_fields["description"].initial += f"\n\n[{t}]"
            if "from_supply" in request.GET:
                pk = request.GET.get("from_supply")
                t = _("Lieferung #%s") % pk
                form.base_fields["name"].initial = t
                form.base_fields["description"].initial += f"\n\n[{t}]"
        return form

    # Save

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        if not change:
            if "from_order" in request.GET:
                pk = request.GET["from_order"]
                if Order.objects.filter(pk=pk).exists():
                    order = Order.objects.get(pk=pk)
                    obj.linked_order = order
                    obj.save()
                    messages.info(
                        request,
                        _("Bestellung #%s wurde mit dieser Notiz verknüpft.") % pk,
                    )
                else:
                    messages.warning(
                        request,
                        _(
                            "Bestellung #%s konnte nicht gefunden werden. Die Notiz wurde trotzdem erstellt."
                        )
                        % pk,
                    )
            if "from_product" in request.GET:
                pk = request.GET["from_product"]
                if Product.objects.filter(pk=pk).exists():
                    product = Product.objects.get(pk=pk)
                    obj.linked_product = product
                    obj.save()
                    messages.info(
                        request, _("Produkt #%s wurde mit dieser Notiz verknüpft.") % pk
                    )
                else:
                    messages.warning(
                        request,
                        _(
                            "Produkt #%s konnte nicht gefunden werden. Die Notiz wurde trotzdem erstellt."
                        )
                        % pk,
                    )
            if "from_customer" in request.GET:
                pk = request.GET["from_customer"]
                if Customer.objects.filter(pk=pk).exists():
                    customer = Customer.objects.get(pk=pk)
                    obj.linked_customer = customer
                    obj.save()
                    messages.info(
                        request, _("Kunde #%s wurde mit dieser Notiz verknüpft.") % pk
                    )
                else:
                    messages.warning(
                        request,
                        _(
                            "Kunde #%s konnte nicht gefunden werden. Die Notiz wurde trotzdem erstellt."
                        )
                        % pk,
                    )
            if "from_supply" in request.GET:
                pk = request.GET["from_supply"]
                if Supply.objects.filter(pk=pk).exists():
                    supply = Supply.objects.get(pk=pk)
                    obj.linked_supply = supply
                    obj.save()
                    messages.info(
                        request,
                        _("Lieferung #%s wurde mit dieser Notiz verknüpft.") % pk,
                    )
                else:
                    messages.warning(
                        request,
                        _(
                            "Lieferung #%s konnte nicht gefunden werden. Die Notiz wurde trotzdem erstellt."
                        )
                        % pk,
                    )


class ProductAdminProductCategoryInline(CustomTabularInline):
    model = Product.categories.through
    extra = 0

    verbose_name = _("Verknüpfte Kategorie")
    verbose_name_plural = _("Verknüpfte Kategorien")

    autocomplete_fields = ("category",)

    # Custom queryset

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related("category", "product")

    # Permissions

    NO_CHANGE = True

    def has_add_permission(self, request, obj=None):
        if not request.user.has_perm("kmuhelper.view_productcategory"):
            return False
        return super().has_add_permission(request, obj)


class ProductAdminChildrenInline(CustomTabularInline):
    model = Product
    extra = 0
    show_change_link = True

    fk_name = "parent"
    verbose_name = _("Untergeordnetes Produkt")
    verbose_name_plural = _("Untergeordnete Produkte")

    fields = [
        "pkfill",
        "article_number",
        "name",
        "selling_price",
    ]

    readonly_fields = ["pkfill", "display_woocommerce_state"]

    def get_fields(self, request, obj=None):
        if is_connected():
            ls = self.fields.copy()
            ls.insert(1, "display_woocommerce_state")
            return ls
        return self.fields

    # Permissions

    NO_ADD = True
    NO_CHANGE = True


@admin.register(Product)
class ProductAdmin(CustomModelAdmin):
    list_display = [
        "pkfill",
        "display_article_number",
        "clean_name",
        "clean_short_description",
        "clean_description",
        "html_image",
        "display_current_price",
        "is_on_sale",
        "linked_note_html",
    ]

    ordering = ("article_number", "name")

    list_filter = (
        ProductTypeFilter,
        WooCommerceStateFilter,
        "supplier",
        "categories",
    )
    search_fields = [
        "pk",
        "article_number",
        "name",
        "short_description",
        "description",
        "note",
        "linked_note__name",
        "linked_note__description",
    ]

    readonly_fields = ["pkfill", "linked_note_html", "display_woocommerce_id"]

    autocomplete_fields = ("supplier",)

    inlines = (ProductAdminProductCategoryInline, ProductAdminChildrenInline)

    save_on_top = True

    list_select_related = ["linked_note"]

    def get_list_display(self, request):
        if is_connected():
            ls = self.list_display.copy()
            ls.insert(1, "display_woocommerce_state")
            return ls
        return self.list_display

    def get_fieldsets(self, request, obj=None):
        fieldsets = [
            (
                _("Verknüpfungen"),
                {
                    "fields": ["display_woocommerce_id", "parent"]
                    if obj and obj.woocommerceid
                    else ["parent"]
                },
            ),
            (_("Infos"), {"fields": ["article_number", "name"]}),
            (
                _("Beschrieb"),
                {
                    "fields": ["short_description", "description"],
                    "classes": ["collapse start-open"],
                },
            ),
            (
                _("Daten"),
                {
                    "fields": [
                        "quantity_description",
                        "selling_price",
                        "vat_rate",
                        "stock_current",
                        "stock_target",
                    ]
                },
            ),
            (
                _("Lieferant"),
                {
                    "fields": [
                        "supplier",
                        "supplier_price",
                        "supplier_article_number",
                        "supplier_url",
                    ],
                    "classes": ["collapse start-open"],
                },
            ),
            (
                _("Aktion"),
                {
                    "fields": ["sale_from", "sale_to", "sale_price"],
                    "classes": ["collapse start-open"],
                },
            ),
            (
                _("Links"),
                {"fields": ["datasheet_url", "image_url"], "classes": ["collapse"]},
            ),
            (
                _("Bemerkung / Notiz"),
                {
                    "fields": ["note", "linked_note_html"] if obj else ["note"],
                    "classes": ["collapse start-open"],
                },
            ),
        ]

        return fieldsets

    # Actions

    @admin.action(
        description=_("Produkte von WooCommerce aktualisieren"), permissions=["change"]
    )
    def wc_update(self, request, queryset):
        WCProductsAPI().bulk_update_objects_from_api(queryset.all(), request)

    @admin.action(description=_("Lagerbestand zurücksetzen"), permissions=["change"])
    def reset_stock(self, request, queryset):
        for product in queryset.all():
            product.stock_current = 0
            product.save()
        count = queryset.count()
        messages.success(
            request,
            ngettext(
                "Lagerbestand von %d Produkt zurückgesetzt.",
                "Lagerbestand von %d Produkten zurückgesetzt.",
                count,
            )
            % count,
        )

    @admin.action(description=_("Aktion beenden"), permissions=["change"])
    def end_sale(self, request, queryset):
        for product in queryset.all():
            product.sale_to = timezone.now()
            product.save()
        count = queryset.count()
        messages.success(
            request,
            ngettext(
                "Aktion von %d Produkt beendet.",
                "Aktion von %d Produkten beendet.",
                count,
            )
            % count,
        )

    actions = ["wc_update", "reset_stock", "end_sale"]

    # Save

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        if obj:
            obj.show_stock_warning(request)


class ProductCategoryAdminChildrenInline(CustomTabularInline):
    model = ProductCategory
    verbose_name = _("Unterkategorie")
    verbose_name_plural = _("Unterkategorien")
    extra = 0

    fields = ("pkfill", "clean_name", "clean_description", "html_image")
    readonly_fields = ("pkfill", "clean_name", "clean_description", "html_image")

    show_change_link = True
    show_full_result_count = True

    # Permissions

    NO_CHANGE = True
    NO_DELETE = True
    NO_ADD = True


class ProductCategoryAdminProductInline(CustomStackedInline):
    model = Product.categories.through
    verbose_name = _("Produkt in dieser Kategorie")
    verbose_name_plural = _("Produkte in dieser Kategorie")
    extra = 0

    autocomplete_fields = ("product",)

    # Custom queryset

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.select_related("category", "product")

    # Permissions

    NO_CHANGE = True

    def has_add_permission(self, request, obj=None):
        if not (
            request.user.has_perm("kmuhelper.view_product")
            or request.user.has_perm("kmuhelper.change_product")
        ):
            return False
        return super().has_add_permission(request, obj)


@admin.register(ProductCategory)
class ProductCategoryAdmin(CustomModelAdmin):
    list_display = [
        "pkfill",
        "clean_name",
        "clean_description",
        "parent_category",
        "html_image",
        "total_quantity",
    ]

    search_fields = ["name", "description"]

    ordering = ("parent_category", "name")

    readonly_fields = ("display_woocommerce_id",)

    inlines = [ProductCategoryAdminChildrenInline, ProductCategoryAdminProductInline]

    list_select_related = ("parent_category",)

    autocomplete_fields = ("parent_category",)

    def get_list_display(self, request):
        if is_connected():
            ls = self.list_display.copy()
            ls.insert(1, "display_woocommerce_state")
            return ls
        return self.list_display

    def get_fieldsets(self, request, obj=None):
        fieldsets = [
            (_("Infos"), {"fields": ["name", "description", "image_url"]}),
            (_("Übergeordnete Kategorie"), {"fields": ["parent_category"]}),
        ]

        if obj and obj.woocommerceid:
            fieldsets.insert(
                0, (_("Verknüpfungen"), {"fields": ["display_woocommerce_id"]})
            )

        return fieldsets

    # Custom queryset

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.annotate(total_quantity=Count("products"))

    # Actions

    @admin.action(
        description=_("Kategorien von WooCommerce aktualisieren"),
        permissions=["change"],
    )
    def wc_update(self, request, queryset):
        WCProductCategoriesAPI().bulk_update_objects_from_api(queryset.all(), request)

    actions = ["wc_update"]


@admin.register(PaymentReceiver)
class PaymentReceiverAdmin(CustomModelAdmin):
    fieldsets = [
        (None, {"fields": ["is_default", "internal_name"]}),
        (
            _("Anzeigeinformationen"),
            {"fields": ["display_name", "display_address_1", "display_address_2"]},
        ),
        (_("Zahlungsinformationen"), {"fields": ["mode", "qriban", "iban"]}),
        (
            _("Rechnungsadresse / Zahlbar an"),
            {
                "fields": [
                    "invoice_name",
                    ("invoice_street", "invoice_street_nr"),
                    ("invoice_postcode", "invoice_city"),
                    "invoice_country",
                ]
            },
        ),
        (
            _("Weitere Darstellungsoptionen"),
            {"fields": ["invoice_display_mode", "swiss_uid", "website", "logourl"]},
        ),
    ]

    list_display = (
        "pkfill",
        "admin_name",
        "mode",
        "active_iban",
        "invoice_display_mode",
        "is_default",
    )
    list_filter = (
        "mode",
        "invoice_country",
        "invoice_display_mode",
    )
    search_fields = [
        "internal_name",
        "display_name",
        "display_address_1",
        "display_address_2",
        "invoice_name",
        "invoice_street",
        "invoice_street_nr",
        "invoice_postcode",
        "invoice_city",
        "invoice_country",
        "swiss_uid",
        "website",
        "logourl",
        "qriban",
        "iban",
    ]

    save_on_top = True

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        if obj.is_default:
            PaymentReceiver.objects.filter(is_default=True).exclude(pk=obj.pk).update(
                is_default=False
            )


modeladmins = [
    (ContactPerson, ContactPersonAdmin),
    (Order, OrderAdmin),
    (ProductCategory, ProductCategoryAdmin),
    (Fee, FeeAdmin),
    (Customer, CustomerAdmin),
    (Supplier, SupplierAdmin),
    (Supply, SupplyAdmin),
    (Note, NoteAdmin),
    (Product, ProductAdmin),
    (PaymentReceiver, PaymentReceiverAdmin),
]



from django.urls import path

from kmuhelper.modules.app import views

#######################

urlpatterns = [
    path("", views.app_index, name="app-index"),
    # Redirects for legacy URLs
    path("manifest.webmanifest", views.manifest_redirect),
    path("mobile/", views.app_redirect),
    path("desktop/", views.app_redirect),
]



from django import forms
from django.utils.translation import gettext_lazy

from kmuhelper.modules.settings.models import Setting
from kmuhelper.modules.settings.utils import set_db_setting

_ = gettext_lazy


class WooCommerceSettingsForm(forms.Form):
    fieldsets = [
        {
            "name": _("WooCommerce-Einstellungen"),
            "fields": ["wc-url", "wc-webhook-secret"],
            "classes": "wide",
        },
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for obj in Setting.objects.filter(id__in=["wc-url", "wc-webhook-secret"]):
            self.fields[obj.id] = obj.get_field()

    def save_settings(self):
        for key, value in self.cleaned_data.items():
            set_db_setting(key, value)



from urllib.parse import urlencode

from django.contrib import admin
from django.db import models
from django.forms.models import model_to_dict
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _, gettext
from kmuhelper import widgets


class CustomModel(models.Model):
    """django.db.models.Model with custom overrides"""

    PKFILL_WIDTH = 6

    ADMIN_TITLE = None
    ADMIN_DESCRIPTION = None
    ADMIN_ICON = "fa-solid fa-circle-exclamation"

    IS_APP_MODEL = False
    NOTE_RELATION = None

    DICT_FIELDS = []
    DICT_EXCLUDE_FIELDS = []

    @admin.display(description="ID", ordering="pk")
    def pkfill(self, width=None):
        """Get the primary key of this object padded with zeros"""

        return str(self.pk).zfill(width or self.PKFILL_WIDTH)

    @property
    def to_dict(self):
        """Return the model fields as a dict"""

        exclude = self.DICT_EXCLUDE_FIELDS
        fields = self.DICT_FIELDS

        if fields:
            return model_to_dict(self, fields=fields, exclude=exclude)

        return model_to_dict(self, exclude=exclude)

    @property
    def to_query_string(self):
        """Return the model fields as a query string"""

        return urlencode(self.to_dict)

    @admin.display(description=_("🔗 Notiz"), ordering="linked_note")
    def linked_note_html(self):
        is_app = self.IS_APP_MODEL
        rel_name = self.NOTE_RELATION

        if rel_name is None:
            return None

        if hasattr(self, "linked_note"):
            if is_app:
                view_name = "admin:kmuhelper_app_todo_change"
            else:
                view_name = "admin:kmuhelper_note_change"

            link = reverse(view_name, kwargs={"object_id": self.linked_note.pk})
            text = gettext("Notiz #%d") % self.linked_note.pk
        else:
            if is_app:
                view_name = "admin:kmuhelper_app_todo_add"
            else:
                view_name = "admin:kmuhelper_note_add"

            link = reverse(view_name) + f"?from_{rel_name}={self.pk}"
            text = gettext("Notiz erstellen")
        return format_html('<a target="_blank" href="{}">{}</a>', link, text)

    class Meta:
        abstract = True


class CustomModelAdmin(admin.ModelAdmin):
    """django.contrib.admin.ModelAdmin with custom overrides"""

    HIDDEN = False

    formfield_overrides = {models.JSONField: {"widget": widgets.PrettyJSONWidget}}
    list_max_show_all = 1000

    def _get_obj_does_not_exist_redirect(self, request, opts, object_id):
        """Redirect to changelist page instead of admin home if object is not found"""

        super()._get_obj_does_not_exist_redirect(request, opts, object_id)
        info = self.model._meta.app_label, self.model._meta.model_name
        return redirect(reverse("admin:%s_%s_changelist" % info))

    def get_actions(self, request):
        """Make some action not always available"""

        actions = super().get_actions(request)
        if "wc_update" in actions:
            from kmuhelper.modules.integrations.woocommerce.utils import is_connected

            if not is_connected():
                del actions["wc_update"]
        return actions

    def has_module_permission(self, request):
        """Add option to hide model in default admin"""

        if self.HIDDEN:
            return {}

        return super().has_module_permission(request)

    # Readonly fields

    def get_readonly_fields(self, request, obj=None) -> tuple:
        if "unlock" in request.GET:
            return tuple(self.readonly_fields)
        return tuple(self.readonly_fields) + tuple(
            self.get_additional_readonly_fields(request, obj)
        )

    def get_additional_readonly_fields(self, request, obj=None) -> list:
        return []

    # Permissions

    NO_CHANGE = False

    def has_change_permission(self, request, obj=None):
        return False if self.NO_CHANGE else super().has_change_permission(request, obj)

    NO_ADD = False

    def has_add_permission(self, request):
        return False if self.NO_ADD else super().has_add_permission(request)

    NO_DELETE = False

    def has_delete_permission(self, request, obj=None):
        return False if self.NO_DELETE else super().has_delete_permission(request, obj)

    NO_VIEW = False

    def has_view_permission(self, request, obj=None):
        return False if self.NO_VIEW else super().has_view_permission(request, obj)


class CustomTabularInline(admin.TabularInline):
    """django.contrib.admin.TabularInline with custom overrides"""

    # Readonly fields

    def get_readonly_fields(self, request, obj=None) -> tuple:
        if "unlock" in request.GET:
            return tuple(self.readonly_fields)
        return tuple(self.readonly_fields) + tuple(
            self.get_additional_readonly_fields(request, obj)
        )

    def get_additional_readonly_fields(self, request, obj=None) -> list:
        return []

    # Permissions

    NO_CHANGE = False

    def has_change_permission(self, request, obj=None):
        return False if self.NO_CHANGE else super().has_change_permission(request, obj)

    NO_ADD = False

    def has_add_permission(self, request, obj=None):
        return False if self.NO_ADD else super().has_add_permission(request, obj)

    NO_DELETE = False

    def has_delete_permission(self, request, obj=None):
        return False if self.NO_DELETE else super().has_delete_permission(request, obj)

    NO_VIEW = False

    def has_view_permission(self, request, obj=None):
        return False if self.NO_VIEW else super().has_view_permission(request, obj)


class CustomStackedInline(admin.StackedInline):
    """django.contrib.admin.StackedInline with custom overrides"""

    # Readonly fields

    def get_readonly_fields(self, request, obj=None) -> tuple:
        if "unlock" in request.GET:
            return tuple(self.readonly_fields)
        return tuple(self.readonly_fields) + tuple(
            self.get_additional_readonly_fields(request, obj)
        )

    def get_additional_readonly_fields(self, request, obj=None) -> list:
        return []

    # Permissions

    NO_CHANGE = False

    def has_change_permission(self, request, obj=None):
        return False if self.NO_CHANGE else super().has_change_permission(request, obj)

    NO_ADD = False

    def has_add_permission(self, request, obj=None):
        return False if self.NO_ADD else super().has_add_permission(request, obj)

    NO_DELETE = False

    def has_delete_permission(self, request, obj=None):
        return False if self.NO_DELETE else super().has_delete_permission(request, obj)

    NO_VIEW = False

    def has_view_permission(self, request, obj=None):
        return False if self.NO_VIEW else super().has_view_permission(request, obj)



import subprocess
import sys

import requests
from django.contrib import messages
from django.shortcuts import render
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from packaging.version import Version, InvalidVersion

from kmuhelper.constants import URL_FAQ


################


def render_error(request, status: int = 404, message: str = ""):
    """Show the error page"""

    if message:
        messages.error(request, message)

    return render(request, "kmuhelper/error.html", status=status)


################


def python_version():
    return str(sys.version.split(" ")[0])


def _package_versions_pypi(project, testpypi=False, allow_prereleases=False):
    url = (
        "https://"
        + ("test." if testpypi else "")
        + "pypi.org/pypi/"
        + project
        + "/json"
    )
    versions = requests.get(url).json()["releases"].keys()
    versions_safe = []
    for v in versions:
        try:
            vs = Version(v)
            if not allow_prereleases and vs.is_prerelease:
                continue
            versions_safe.append(vs)
        except InvalidVersion:
            continue
    return sorted(versions_safe)


def _package_version_local(package) -> Version:
    cmd = [sys.executable, "-m", "pip", "show", "{}".format(package)]
    ver = str(subprocess.run(cmd, capture_output=True, text=True))
    ver = ver[ver.find("Version:") + 8 :]
    ver = ver[: ver.find("\\n")].replace(" ", "")
    try:
        return Version(ver)
    except InvalidVersion:
        return None


def package_version(package, testpypi=False):
    all_versions = _package_versions_pypi(package, testpypi)
    latest_version = all_versions[-1] if all_versions else None
    current_version = _package_version_local(package)

    if latest_version and current_version:
        uptodate = latest_version <= current_version
    else:
        uptodate = None

    return {
        # "all":      all_versions,
        "latest": str(latest_version),
        "current": str(current_version),
        "uptodate": uptodate,
    }


################


def getfirstindex(data: list, search: list):
    """Get the index of the first occurence of any string from search in data"""
    for s in search:
        if s in data:
            return data.index(s)
    return None


def formatprice(price):
    """Format a float to exactly 2 decimal places"""
    return "{:.2f}".format(float(price))


def runden(price, to=0.05):
    """Round a float to .05 or any other value"""
    # Note: formatprice is used because of floating point approximation
    return float(formatprice(float(round(round(price / to) * to, 2))))


###############


def modulo10rekursiv(strNummer):
    intTabelle = [
        [0, 9, 4, 6, 8, 2, 7, 1, 3, 5],
        [9, 4, 6, 8, 2, 7, 1, 3, 5, 0],
        [4, 6, 8, 2, 7, 1, 3, 5, 0, 9],
        [6, 8, 2, 7, 1, 3, 5, 0, 9, 4],
        [8, 2, 7, 1, 3, 5, 0, 9, 4, 6],
        [2, 7, 1, 3, 5, 0, 9, 4, 6, 8],
        [7, 1, 3, 5, 0, 9, 4, 6, 8, 2],
        [1, 3, 5, 0, 9, 4, 6, 8, 2, 7],
        [3, 5, 0, 9, 4, 6, 8, 2, 7, 1],
        [5, 0, 9, 4, 6, 8, 2, 7, 1, 3],
    ]
    strNummer = strNummer.replace(" ", "")
    uebertrag = 0
    for num in strNummer:
        uebertrag = intTabelle[uebertrag][int(num)]
    return [0, 9, 8, 7, 6, 5, 4, 3, 2, 1][uebertrag]


###############


def faq(anchor="", text=_("FAQ")):
    """Get a link to the FAQ page in the manual."""

    href = f"{URL_FAQ}#{anchor}"
    return format_html('<a target="_blank" href="{}">{}</a>', href, text)



from io import BytesIO

from django.core import mail
from django.template.loader import get_template

# These utils are not currently used, but they are still here for reference


def send_mail(subject: str, to: str, template_name: str, context: dict = {}, **kwargs):
    html_message = get_template("kmuhelper/emails/" + template_name).render(context)

    msg = mail.EmailMessage(subject=subject, body=html_message, to=[to], **kwargs)

    msg.content_subtype = "html"

    return bool(msg.send())


def send_pdf(
    subject: str,
    to: str,
    template_name: str,
    pdf: BytesIO,
    pdf_filename: str = "file.pdf",
    context: dict = {},
    **kwargs
):
    html_message = get_template("kmuhelper/emails/" + template_name).render(context)

    msg = mail.EmailMessage(subject=subject, body=html_message, to=[to], **kwargs)

    msg.content_subtype = "html"
    msg.attach(filename=pdf_filename, content=pdf.read(), mimetype="application/pdf")

    return bool(msg.send())



# Generated by Django 4.2.7 on 2023-11-22 22:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("kmuhelper", "0106_alter_app_arrival_options_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="emailtemplate",
            name="mail_subject",
            field=models.CharField(
                help_text="Unterstützt Platzhalter. Siehe unten für mehr Informationen.",
                max_length=50,
                verbose_name="Betreff",
            ),
        ),
        migrations.AlterField(
            model_name="emailtemplate",
            name="mail_text",
            field=models.TextField(
                help_text="Unterstützt Platzhalter. Siehe unten für mehr Informationen.",
                verbose_name="Text",
            ),
        ),
        migrations.AlterField(
            model_name="emailtemplate",
            name="mail_to",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Unterstützt Platzhalter. Siehe unten für mehr Informationen.",
                max_length=50,
                verbose_name="Empfänger",
            ),
        ),
    ]



"""
Tests for the product model
"""

import unittest
from datetime import timedelta

from django.utils import timezone

from kmuhelper.modules.main.models import Product


class ProductTest(unittest.TestCase):
    def test_sale_price_ended_before(self):
        sale_from = timezone.now() - timedelta(days=7)
        sale_to = timezone.now() - timedelta(seconds=1)
        prod = Product(
            selling_price=10, sale_price=5, sale_from=sale_from, sale_to=sale_to
        )
        self.assertAlmostEqual(prod.get_current_price(), 10)
        self.assertEqual(prod.display_current_price(), "10.00 CHF")

    def test_sale_price_starts_after(self):
        sale_from = timezone.now() + timedelta(seconds=10)
        sale_to = timezone.now() + timedelta(days=7)
        prod = Product(
            selling_price=10, sale_price=5, sale_from=sale_from, sale_to=sale_to
        )
        self.assertAlmostEqual(prod.get_current_price(), 10)
        self.assertEqual(prod.display_current_price(), "10.00 CHF")

    def test_sale_price_current(self):
        sale_from = timezone.now() - timedelta(seconds=10)
        sale_to = timezone.now() + timedelta(seconds=10)
        prod = Product(
            selling_price=10, sale_price=5, sale_from=sale_from, sale_to=sale_to
        )
        self.assertAlmostEqual(prod.get_current_price(), 5)
        self.assertEqual(prod.display_current_price(), "5.00 CHF")



# Generated by Django 5.0 on 2024-02-01 16:37

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("kmuhelper", "0110_orderitem_made_product_optional"),
    ]

    operations = [
        migrations.AlterField(
            model_name="order",
            name="tracking_number",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Trackingnummer ohne Leerzeichen. (optional)",
                max_length=35,
                validators=[
                    django.core.validators.RegexValidator(
                        "^[A-Z0-9\\.]{8,}$",
                        "Erlaubte Zeichen: Grossbuchstaben, Zahlen und Punkte. Länge: 8-35 Zeichen",
                    )
                ],
                verbose_name="Trackingnummer",
            ),
        ),
    ]



from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.shortcuts import redirect, render
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy, gettext

import kmuhelper.modules.config as config
from kmuhelper.decorators import (
    require_object,
    confirm_action,
    require_all_kmuhelper_perms,
    require_kmuhelper_module_perms,
)
from kmuhelper.modules.emails.models import EMail, Attachment, EMailTemplate
from kmuhelper.utils import render_error

_ = gettext_lazy

#####

# EMail views


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["send_email"])
@require_object(EMail)
@confirm_action(_("E-Mail-Nachricht senden"))
def email_send(request, obj):
    if obj.is_valid():
        try:
            success = obj.send()
            if success:
                messages.success(request, gettext("E-Mail wurde versendet!"))
            else:
                messages.error(request, gettext("E-Mail konnte nicht gesendet werden!"))
        except FileNotFoundError:
            messages.error(
                request,
                gettext("Eine Datei aus dem Anhang konnte nicht gefunden werden!"),
            )
    return redirect(reverse("admin:kmuhelper_email_change", args=[obj.pk]))


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["send_email"])
@require_object(EMail)
@confirm_action(_("E-Mail Nachricht ERNEUT senden"))
def email_resend(request, obj):
    if obj.is_valid(request):
        try:
            success = obj.send()
            if success:
                messages.success(request, gettext("E-Mail wurde erneut versendet!"))
            else:
                messages.error(request, gettext("E-Mail konnte nicht gesendet werden!"))
        except FileNotFoundError:
            messages.error(
                request,
                gettext("Eine Datei aus dem Anhang konnte nicht gefunden werden!"),
            )
    return redirect(reverse("admin:kmuhelper_email_change", args=[obj.pk]))


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["view_email"])
@require_object(EMail)
def email_preview(request, obj):
    if obj.is_valid(request):
        return HttpResponse(obj.render(online=True))
    return redirect(reverse("admin:kmuhelper_email_change", args=[obj.pk]))


# Public views


@require_object(EMail, show_errorpage=True)
def email_view(request, obj):
    """Render an email for online viewing"""

    token_received = request.GET.get("token", None)
    token_stored = str(obj.token)

    if request.user.has_perm("kmuhelper.view_email") or (
        token_received == token_stored
    ):
        if obj.is_valid():
            return HttpResponse(obj.render(online=True))

        messages.warning(
            request, gettext("Diese E-Mail kann zurzeit leider nicht angezeigt werden.")
        )
    else:
        messages.error(request, gettext("Du hast keinen Zugriff auf diese E-Mail!"))

    return render_error(request)


# EMailAttachment views


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["download_attachment"])
@require_object(Attachment)
def attachment_download(request, obj):
    download = not "preview" in request.GET
    try:
        return obj.get_file_response(download=download)
    except FileNotFoundError:
        messages.error(request, gettext("Datei wurde nicht gefunden!"))
        return render_error(request)


# Public views


@require_object(Attachment, show_errorpage=True)
def attachment_view(request, obj):
    """Render an attachment for online viewing"""

    token_received = request.GET.get("token", None)
    token_stored = str(obj.token)

    if request.user.has_perm("kmuhelper.download_attachment") or (
        token_received == token_stored
    ):
        download = "download" in request.GET
        try:
            return obj.get_file_response(download=download)
        except FileNotFoundError:
            messages.error(
                request, gettext("Diese Datei ist leider nicht mehr verfügbar!")
            )
    else:
        messages.error(
            request, gettext("Du hast keinen Zugriff auf diesen E-Mail-Anhang!")
        )

    return render_error(request)


# EMailTemplate Views


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["view_emailtemplate"])
def emailtemplate_savevars(request):
    """Save variables into session for usage in template"""

    data = request.GET.dict()
    request.session["kmuhelper_emailtemplate_vars"] = data

    messages.success(
        request,
        gettext(
            "Daten wurden erfolgreich zwischengespeichert, um die nächste Vorlage auszufüllen."
        ),
    )
    return redirect(reverse("admin:kmuhelper_emailtemplate_changelist"))


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["view_emailtemplate"])
def emailtemplate_resetvars(request):
    """Reset saved session variables for email templates"""

    if "kmuhelper_emailtemplate_vars" in request.session:
        del request.session["kmuhelper_emailtemplate_vars"]

    messages.success(request, _("Gespeicherte Daten wurden gelöscht!"))
    return redirect(reverse("admin:kmuhelper_emailtemplate_changelist"))


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["view_emailtemplate", "add_email"])
@require_object(EMailTemplate)
def emailtemplate_use(request, obj):
    """Use a template"""

    savedvars = request.session.pop("kmuhelper_emailtemplate_vars", dict())
    requestvars = request.GET.dict()
    data = {**savedvars, **requestvars}

    mail = obj.create_mail(data)

    messages.success(
        request, _("Vorlage wurde mit folgenden Daten ausgefüllt: %s") % data
    )
    return redirect(reverse("admin:kmuhelper_email_change", args=[mail.pk]))


# Custom fake admin


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_kmuhelper_module_perms("emails")
def email_index(request):
    return render(
        request,
        "admin/kmuhelper/_special/email-index.html",
        config.get_module_home_context(request, "emails"),
    )



import kmuhelper.modules.config as config
from django.conf import settings
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import render, redirect
from django.template import TemplateDoesNotExist, TemplateSyntaxError
from django.template.loader import get_template
from django.urls import NoReverseMatch, reverse_lazy, reverse
from django.utils.translation import gettext
from kmuhelper.constants import URL_MANUAL
from kmuhelper.decorators import require_any_kmuhelper_perms
from kmuhelper.utils import render_error

_ = gettext

# Create your views here.


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_any_kmuhelper_perms()
def home(request):
    grid = [
        [
            {
                "title": _("Admin"),
                "subtitle": _("Volle Kontrolle über die Hauptfunktionen"),
                "url": reverse("admin:app_list", kwargs={"app_label": "kmuhelper"}),
                "icon": "fa-solid fa-key",
                "hidden": not config.user_has_module_permission(request.user, "main"),
            },
            {
                "title": _("App"),
                "subtitle": _(
                    "Eingeschränkte Verwaltung von Daten in verschiedenen Arbeitsschritten"
                ),
                "url": reverse("kmuhelper:app-index"),
                "icon": "fa-solid fa-desktop",
                "hidden": not config.user_has_module_permission(request.user, "app"),
            },
        ],
        [
            {
                "title": _("E-Mails"),
                "subtitle": _("Verwaltung und Erstellung von E-Mails"),
                "url": reverse("kmuhelper:email-index"),
                "icon": "fa-solid fa-envelope-open",
                "hidden": not config.user_has_module_permission(request.user, "emails"),
            },
            {
                "title": _("Statistiken"),
                "subtitle": _("Diagramme und co."),
                "url": reverse("kmuhelper:stats"),
                "icon": "fa-solid fa-chart-pie",
                "hidden": not config.user_has_module_permission(request.user, "main"),
            },
        ],
        [
            {
                "title": _("Einstellungen"),
                "subtitle": _("Einstellungen für den KMUHelper"),
                "url": reverse("kmuhelper:settings"),
                "icon": "fa-solid fa-cog",
                "hidden": not config.user_has_module_permission(
                    request.user, "settings"
                ),
            },
            {
                "title": _("Handbuch"),
                "subtitle": _("Dokumentation zum KMUHelper"),
                "url": URL_MANUAL,
                "icon": "fa-solid fa-book",
            },
        ],
    ]
    return render(
        request, "kmuhelper/home.html", {"has_permission": True, "grid": grid}
    )


def manifest(request):
    response = render(request, "kmuhelper/manifest.json", {})
    response["Content-Type"] = "text/json"
    response["Service-Worker-Allowed"] = "/"
    return response


def login(request):
    search = request.GET.urlencode()
    url = settings.LOGIN_URL or reverse("admin:login")
    return redirect(url + ("?" + search if search else ""))


@require_any_kmuhelper_perms()
def error(request):
    return render_error(request)


@permission_required("is_superuser")
def _templatetest(request, templatename):
    """DEBUG: Test a template with given parameters"""

    try:
        get_template(templatename)
        return render(request, templatename, request.GET.dict())
    except TemplateDoesNotExist:
        return render_error(
            request,
            status=404,
            message=_("Vorlage %s wurde nicht gefunden.") % templatename,
        )
    except TemplateSyntaxError as error:
        return render_error(
            request,
            status=400,
            message=_("Vorlage %(name)s enthält einen Syntaxfehler: %(error)s")
            % {"name": templatename, "error": error},
        )
    except NoReverseMatch as error:
        return render_error(
            request, status=400, message=_("Es ist ein Fehler aufgetreten: %s") % error
        )



from django.urls import path

from kmuhelper.modules.settings import views

#######################

urlpatterns = [
    path("", views.settings_form, name="settings"),
    path("build-info", views.build_info, name="build-info"),
]



# Generated by Django 4.1.6 on 2023-02-23 16:26

from django.db import migrations, models


def copy_content_to_new_fields(apps, schema_editor):
    PaymentReceiver = apps.get_model("kmuhelper", "PaymentReceiver")
    for receiver in PaymentReceiver.objects.all():
        receiver.internal_name = receiver.invoice_name
        receiver.display_name = receiver.invoice_name
        receiver.display_address_1 = receiver.invoice_address_1
        receiver.display_address_2 = receiver.invoice_address_2
        receiver.save()


class Migration(migrations.Migration):
    dependencies = [
        ("kmuhelper", "0101_squashed_0100_to_0100"),
    ]

    operations = [
        migrations.RenameField(
            model_name="paymentreceiver",
            old_name="address_1",
            new_name="invoice_address_1",
        ),
        migrations.RenameField(
            model_name="paymentreceiver",
            old_name="address_2",
            new_name="invoice_address_2",
        ),
        migrations.RenameField(
            model_name="paymentreceiver",
            old_name="country",
            new_name="invoice_country",
        ),
        migrations.RenameField(
            model_name="paymentreceiver",
            old_name="name",
            new_name="invoice_name",
        ),
        migrations.RemoveField(
            model_name="paymentreceiver",
            name="email",
        ),
        migrations.RemoveField(
            model_name="paymentreceiver",
            name="phone",
        ),
        migrations.AddField(
            model_name="paymentreceiver",
            name="display_address_1",
            field=models.CharField(
                default="",
                help_text="Wird oben auf der Rechnung angezeigt.",
                max_length=70,
                verbose_name="Adresszeile 1",
            ),
        ),
        migrations.AddField(
            model_name="paymentreceiver",
            name="display_address_2",
            field=models.CharField(
                default="",
                help_text="Wird oben auf der Rechnung angezeigt.",
                max_length=70,
                verbose_name="Adresszeile 2",
            ),
        ),
        migrations.AddField(
            model_name="paymentreceiver",
            name="display_name",
            field=models.CharField(
                default="",
                help_text="Wird oben auf der Rechnung angezeigt.",
                max_length=70,
                verbose_name="Anzeigename",
            ),
        ),
        migrations.AddField(
            model_name="paymentreceiver",
            name="internal_name",
            field=models.CharField(
                default="", max_length=250, verbose_name="Interne Bezeichnung"
            ),
        ),
        migrations.AlterField(
            model_name="paymentreceiver",
            name="invoice_address_1",
            field=models.CharField(
                help_text="In QR-Rechnung 'Zahlbar an' - Strasse und Hausnummer oder 'Postfach'",
                max_length=70,
                verbose_name="Adresszeile 1",
            ),
        ),
        migrations.AlterField(
            model_name="paymentreceiver",
            name="invoice_address_2",
            field=models.CharField(
                help_text="In QR-Rechnung 'Zahlbar an' - PLZ und Ort",
                max_length=70,
                verbose_name="Adresszeile 2",
            ),
        ),
        migrations.AlterField(
            model_name="paymentreceiver",
            name="invoice_country",
            field=models.CharField(
                choices=[("CH", "Schweiz"), ("LI", "Liechtenstein")],
                default="CH",
                help_text="In QR-Rechnung 'Zahlbar an'",
                max_length=2,
                verbose_name="Land",
            ),
        ),
        migrations.AlterField(
            model_name="paymentreceiver",
            name="invoice_name",
            field=models.CharField(
                help_text="In QR-Rechnung 'Zahlbar an' - Kontoinhaber / Firma",
                max_length=70,
                verbose_name="Name",
            ),
        ),
        migrations.RunPython(
            code=copy_content_to_new_fields,
            reverse_code=migrations.RunPython.noop,
            elidable=True,
        ),
    ]



from django.forms.widgets import Textarea
from django.core.exceptions import ValidationError
from django.core import validators

MULTI_EMAIL_FIELD_EMPTY_VALUES = validators.EMPTY_VALUES + ("[]",)

try:
    from django.utils import six

    string_types = six.string_types
except ImportError:
    string_types = str


class MultiEmailWidget(Textarea):
    is_hidden = False

    def prep_value(self, value):
        """Prepare value before effectively render widget"""
        if value in MULTI_EMAIL_FIELD_EMPTY_VALUES:
            return ""
        elif isinstance(value, string_types):
            return value
        elif isinstance(value, list):
            return "\n".join(value)
        raise ValidationError("Invalid format.")

    def render(self, name, value, **kwargs):
        value = self.prep_value(value)
        return super(MultiEmailWidget, self).render(name, value, **kwargs)



# Generated by Django 5.0.4 on 2024-04-09 14:22

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("kmuhelper", "0112_product_parent"),
    ]

    operations = [
        migrations.AddField(
            model_name="customer",
            name="woocommerce_deleted",
            field=models.BooleanField(
                blank=True, default=False, verbose_name="Deleted in WooCommerce?"
            ),
        ),
        migrations.AddField(
            model_name="order",
            name="woocommerce_deleted",
            field=models.BooleanField(
                blank=True, default=False, verbose_name="Deleted in WooCommerce?"
            ),
        ),
        migrations.AddField(
            model_name="product",
            name="woocommerce_deleted",
            field=models.BooleanField(
                blank=True, default=False, verbose_name="Deleted in WooCommerce?"
            ),
        ),
        migrations.AddField(
            model_name="productcategory",
            name="woocommerce_deleted",
            field=models.BooleanField(
                blank=True, default=False, verbose_name="Deleted in WooCommerce?"
            ),
        ),
    ]



# Generated by Django 5.0.4 on 2024-04-17 23:03

from django.db import migrations, models


def migrate_address_fields(apps, schema_editor):
    PaymentReceiver = apps.get_model("kmuhelper", "PaymentReceiver")
    for receiver in PaymentReceiver.objects.all():
        receiver.invoice_street = receiver.invoice_address_1[:70]

        _index = receiver.invoice_address_2.index(" ")
        receiver.invoice_postcode = receiver.invoice_address_2[:_index]
        receiver.invoice_city = receiver.invoice_address_2[_index + 1 : _index + 36]

        receiver.save()


def migrate_address_fields_reverse(apps, schema_editor):
    PaymentReceiver = apps.get_model("kmuhelper", "PaymentReceiver")
    for receiver in PaymentReceiver.objects.all():
        receiver.invoice_address_1 = (
            f"{receiver.invoice_street} {receiver.invoice_street_nr}".strip()
        )
        receiver.invoice_address_2 = (
            f"{receiver.invoice_postcode} {receiver.invoice_city}".strip()
        )
        receiver.save()


class Migration(migrations.Migration):
    dependencies = [
        ("kmuhelper", "0114_adminlogentry_alter_order_payment_conditions"),
    ]

    operations = [
        migrations.AddField(
            model_name="paymentreceiver",
            name="invoice_city",
            field=models.CharField(default="", max_length=35, verbose_name="Ort"),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="paymentreceiver",
            name="invoice_postcode",
            field=models.CharField(
                default="", max_length=16, verbose_name="Postleitzahl"
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="paymentreceiver",
            name="invoice_street",
            field=models.CharField(
                default="",
                help_text="Strasse oder Postfach",
                max_length=70,
                verbose_name="Strasse",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="paymentreceiver",
            name="invoice_street_nr",
            field=models.CharField(
                blank=True, default="", max_length=16, verbose_name="Hausnummer"
            ),
            preserve_default=False,
        ),
        migrations.RunPython(
            code=migrate_address_fields,
            reverse_code=migrate_address_fields_reverse,
            elidable=True,
        ),
        migrations.RemoveField(
            model_name="paymentreceiver",
            name="invoice_address_1",
        ),
        migrations.RemoveField(
            model_name="paymentreceiver",
            name="invoice_address_2",
        ),
        migrations.AlterField(
            model_name="paymentreceiver",
            name="invoice_country",
            field=models.CharField(
                choices=[("CH", "Schweiz"), ("LI", "Liechtenstein")],
                default="CH",
                max_length=2,
                verbose_name="Land",
            ),
        ),
        migrations.AlterField(
            model_name="paymentreceiver",
            name="invoice_name",
            field=models.CharField(
                help_text="Name / Firma", max_length=70, verbose_name="Kontoinhaber"
            ),
        ),
    ]



from datetime import datetime

from django.contrib.admin.models import LogEntry, CHANGE
from django.contrib.auth.decorators import login_required
from django.shortcuts import redirect, render
from django.urls import reverse_lazy, reverse
from django.utils import translation, timezone

from kmuhelper import constants
from kmuhelper.decorators import require_object, require_all_kmuhelper_perms
from kmuhelper.modules.main.models import Order
from kmuhelper.modules.pdfgeneration.order.forms import PDFOrderForm
from kmuhelper.modules.pdfgeneration.order.generator import PDFOrder
from kmuhelper.translations import Language
from kmuhelper.utils import render_error

_ = translation.gettext

# Views


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["view_order"])
@require_object(Order)
def order_view_pdf(request, obj):
    order = obj
    is_print_version = "print" in request.GET
    is_download = "download" in request.GET

    preset = request.GET.get("preset", "invoice")
    title = None
    text = None

    if "custom" in request.GET:
        title = order.pdf_title
        text = order.pdf_text

    # Get the language
    defaultlang = order.language
    lang = request.GET.get("language", defaultlang).lower()
    if lang not in dict(constants.LANGUAGES):
        lang = defaultlang

    # Set invoice date if not present
    if order.invoice_date is None:
        order.invoice_date = timezone.now()
        order.save()

    with Language(lang):
        match preset:
            case "invoice":
                title = title or _("Rechnung")
                pdf = PDFOrder(
                    order,
                    title,
                    text=text,
                    lang=lang,
                    add_cut_lines=not is_print_version,
                )
            case "delivery-note":
                title = title or _("Lieferschein")
                pdf = PDFOrder(
                    order, title, text=text, lang=lang, is_delivery_note=True
                )
            case "payment-reminder":
                days_to_pay = 14
                title = title or _("Zahlungserinnerung")
                text = text or _(
                    "Geschätzter Kunde\n\nVermutlich ist Ihnen entgangen, diese Rechnung vom %(original_date)s innert "
                    "der gewährten Frist zu begleichen.\nWir bitten Sie, dies innert %(days)d Tagen "
                    "nachzuholen.\n\nSollte sich Ihre Zahlung mit diesem Schreiben überkreuzen, so betrachten Sie "
                    "dieses bitte als gegenstandslos."
                ) % {
                    "original_date": order.invoice_date.strftime("%d.%m.%Y"),
                    "days": days_to_pay,
                }
                # Temporarily replace the order's date and payment conditions
                order.invoice_date = datetime.now().date()
                order.payment_conditions = f"0:{days_to_pay}"
                pdf = PDFOrder(
                    order,
                    title,
                    text=text,
                    lang=lang,
                    add_cut_lines=not is_print_version,
                    show_payment_conditions=False,
                )
            case _:
                return render_error(
                    request, status=400, message="Ungültige Vorlage: " + str(preset)
                )

        # Log the action
        LogEntry.objects.log_actions(
            user_id=request.user.id,
            queryset=[obj],
            action_flag=CHANGE,
            change_message=f'PDF "{title}" aus Vorlage "{preset}" erstellt mit Text: "{text}"'
            if text
            else f'PDF "{title}" aus Vorlage "{preset}" erstellt.',
        )

        filename = f"{str(order)} - {title}.pdf"
        return pdf.get_response(as_attachment=is_download, filename=filename)


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["view_order"])
@require_object(Order)
def order_create_pdf_form(request, obj):
    initial = {
        "language": obj.language,
        "title": obj.pdf_title,
        "text": obj.pdf_text,
    }

    if request.method == "POST":
        form = PDFOrderForm(request.POST)
        if form.is_valid():
            form.update_order_settings(obj)
            url = reverse("admin:kmuhelper_order_pdf", args=[obj.id])
            return redirect(url + form.get_url_params())
    else:
        form = PDFOrderForm(initial=initial)

    return render(
        request,
        "admin/kmuhelper/order/pdf_form.html",
        {
            "form": form,
            "has_permission": True,
            "original": obj,
        },
    )
